#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------


"""This module contains the tools for autoupdating ipfs hashes in the documentation."""

import argparse
import itertools
import re
import sys
from pathlib import Path
from typing import Dict, List, Optional

import yaml
from aea.cli.packages import get_package_manager
from aea.configurations.data_types import PackageId
from aea.helpers.base import IPFS_HASH_REGEX, SIMPLE_ID_REGEX


CLI_REGEX = r"(?P<cli>aea|autonomy)"
# CMD_REGEX should be r"(?P<cmd>(\S+\s(\s--\S+)*)+)",
# but python implementation differs from others and does not match it properly
CMD_REGEX = r"(?P<cmd>.*)"
VENDOR_REGEX = rf"(?P<vendor>{SIMPLE_ID_REGEX})"
PACKAGE_REGEX = rf"(?P<package>{SIMPLE_ID_REGEX})"
VERSION_REGEX = r"(?P<version>\d+\.\d+\.\d+)"
FLAGS_REGEX = r"(?P<flags>(\s--.*)?)"
PACKAGE_TYPE_REGEX = (
    r"(?P<package_type>(skill|protocol|connection|contract|agent|service))"
)

AEA_COMMAND_REGEX = rf"(?P<full_cmd>{CLI_REGEX} {CMD_REGEX} (?:{VENDOR_REGEX}\/{PACKAGE_REGEX}:{VERSION_REGEX}?:?)?(?P<hash>{IPFS_HASH_REGEX}){FLAGS_REGEX})"
FULL_PACKAGE_REGEX = rf"(?P<full_package>(?:{VENDOR_REGEX}\/{PACKAGE_REGEX}:{VERSION_REGEX}?:?)?(?P<hash>{IPFS_HASH_REGEX}))"
PACKAGE_TABLE_REGEX = rf"\|\s*{PACKAGE_TYPE_REGEX}\/{VENDOR_REGEX}\/{PACKAGE_REGEX}\/{VERSION_REGEX}\s*\|\s*`(?P<hash>{IPFS_HASH_REGEX})`\s*\|"
PACKAGE_MAPPING_REGEX = rf"(?P<package_mapping>(?:\"{PACKAGE_TYPE_REGEX}\/{VENDOR_REGEX}\/{PACKAGE_REGEX}\/{VERSION_REGEX}\":)\s*\"(?P<hash>{IPFS_HASH_REGEX})\")"

ROOT_DIR = Path(__file__).parent.parent
HASH_SKIPS = [
    "Qmbh9SQLbNRawh9Km3PMEDSxo77k1wib8fYZUdZkhPBiev",  # Testing image used in tutorials/examples.
    "bafybei0000000000000000000000000000000000000000000000000000",  # Placeholder hash used in tutorials/examples.
]


def read_file(filepath: str) -> str:
    """Loads a file into a string"""
    with open(filepath, "r", encoding="utf-8") as file_:
        file_str = file_.read()
    return file_str


def get_packages() -> Dict[str, str]:
    """Get packages."""
    data = get_package_manager(Path("packages").relative_to(".")).json
    if "dev" in data:
        return {**data["dev"], **data["third_party"]}
    return data


class Package:  # pylint: disable=too-few-public-methods
    """Class that represents a package in packages.json"""

    def __init__(self, package_id_str: str, package_hash: str) -> None:
        """Constructor"""

        self.package_id = PackageId.from_uri_path(package_id_str)
        self.vendor = self.package_id.author
        self.type = self.package_id.package_type.to_plural()
        self.name = self.package_id.name
        self.hash = package_hash

        if self.name == "scaffold":
            return

        if self.type not in (
            "connections",
            "agents",
            "protocols",
            "services",
            "skills",
            "contracts",
        ):
            raise ValueError(
                f"Package: unknown package type in packages.json: {self.type}"
            )
        self.type = self.type[:-1]  # remove last s

        self.last_version = None
        yaml_file_path = Path(
            ROOT_DIR,
            "packages",
            self.vendor,
            self.type + "s",
            self.name,
            f"{'aea-config' if self.type == 'agent' else self.type}.yaml",
        )
        with open(yaml_file_path, "r", encoding="utf-8") as file:
            content = yaml.load_all(file, Loader=yaml.FullLoader)
            for resource in content:
                if "version" in resource:
                    self.last_version = resource["version"]
                    break

    def get_command(
        self, cmd: str, include_version: bool = True, flags: str = ""
    ) -> str:
        """
        Get the corresponding command.

        :param cmd: the command
        :param include_version: whether or not to include the version
        :param flags: command flags
        :return: the full command
        """
        version = (
            ":" + self.last_version if include_version and self.last_version else ""
        )
        return f"autonomy {cmd} {self.vendor}/{self.name}{version}:{self.hash}{flags}"


class PackageHashManager:
    """Class that represents the packages in packages.json"""

    def __init__(self) -> None:
        """Constructor"""
        packages = get_packages()
        self.packages = [Package(key, value) for key, value in packages.items()]

        self.package_tree: Dict = {}
        for p in self.packages:
            self.package_tree.setdefault(p.vendor, {})
            self.package_tree[p.vendor].setdefault(p.type, {})
            self.package_tree[p.vendor][p.type].setdefault(p.name, p)
            assert re.match(IPFS_HASH_REGEX, p.hash)  # detect wrong regexes

    def get_package_by_hash(self, package_hash: str) -> Optional[Package]:
        """Get a package given its hash"""
        packages = list(filter(lambda p: p.hash == package_hash, self.packages))
        if not packages:
            return None
        if len(packages) > 1:
            raise ValueError(
                f"PackageHashManager: hash search for {package_hash} returned more than 1 result in packages.json"
            )
        return packages[0]

    def get_hash_by_package_line(
        self, package_line: str, target_file: str
    ) -> Optional[str]:
        """Get a hash given its package line"""

        try:
            m_command = re.match(AEA_COMMAND_REGEX, package_line)
            m_package = re.match(FULL_PACKAGE_REGEX, package_line)

            # No match
            if not m_command and not m_package:
                print(
                    f"[{target_file}]: line '{package_line}' does not match an autonomy/aea command or package format"
                )
                return None
            m = m_command or m_package
            d = m.groupdict()  # type: ignore

            # Underspecified commands that only use the hash
            # In this case we cannot infer the package type, just check whether or not the hash exists in packages.json
            if not d["vendor"] and not d["package"]:
                package = self.get_package_by_hash(d["hash"])

                # This hash exists in packages.json
                if package:
                    return package.hash

                # This hash does not exist in packages.json
                print(
                    f"[{target_file}]: unknown IPFS hash in line '{package_line}'. Can't fix because this command just uses the hash"
                )
                return None

            # Complete command, succesfully retrieved or complete packages

            # Guess the package type (agent, service, contract...). First try to find the package in the package_tree
            potential_package_types = []
            for package_type, packages in self.package_tree[d["vendor"]].items():
                if d["package"] in packages.keys():
                    potential_package_types.append(package_type)

            # If only 1 match has been found we can be sure about the package type
            if len(potential_package_types) == 1:
                package_type = potential_package_types[0]
            else:
                # Try to guess the package type from the command
                package_type = None

                # Fetch option is only available for agents and services
                if d["cmd"] == "fetch":
                    package_type = (
                        "service" if "--service" in d["full_cmd"] else "agent"
                    )

                # Deployments are always services
                if "deployment" in d["cmd"]:
                    package_type = "service"

                # Add commands always specify the package type
                if d["cmd"].startswith("add"):
                    package_type = d["cmd"].split(" ")[-1]  # i.e.: aea add connection

                if not package_type:
                    raise ValueError(
                        f"[{target_file}]: could not infer the package type for line '{package_line}'\nPlease update the hash manually."
                    )

            return self.package_tree[d["vendor"]][package_type][d["package"]].hash

        # Otherwise log the error
        except KeyError:
            print(
                f"[{target_file}]: could not find the corresponding hash for line '{package_line}'"
            )
            return None

    def get_hash_by_attributes(
        self, package_type: str, vendor: str, package_name: str
    ) -> str:
        """Get a package hash give the package information"""
        return self.package_tree[vendor][package_type][package_name].hash


def check_ipfs_hashes(  # pylint: disable=too-many-locals,too-many-statements
    paths: Optional[List[Path]] = None, fix: bool = False
) -> None:
    """Fix ipfs hashes in the docs"""

    paths = paths or [Path("docs")]

    all_md_files = itertools.chain.from_iterable([path.rglob("*.md") for path in paths])
    errors = False
    hash_mismatches = False
    old_to_new_hashes = {}
    package_manager = PackageHashManager()
    matches = 0

    # Fix hashes in docs
    for md_file in all_md_files:
        content = read_file(str(md_file))

        # Fix full commands in docs
        for match in [m.groupdict() for m in re.finditer(AEA_COMMAND_REGEX, content)]:
            matches += 1
            doc_full_cmd = match["full_cmd"]
            doc_cmd = match["cmd"]
            doc_hash = match["hash"]
            flags = match["flags"]

            if doc_hash in HASH_SKIPS:
                continue

            expected_hash = package_manager.get_hash_by_package_line(
                doc_full_cmd, str(md_file)
            )
            if not expected_hash:
                errors = True
                continue
            expected_package = package_manager.get_package_by_hash(expected_hash)
            if not expected_package:
                errors = True
                continue

            new_command = expected_package.get_command(cmd=doc_cmd, flags=flags)

            # Overwrite with new hash
            if doc_hash == expected_hash:
                continue

            hash_mismatches = True

            if fix:
                content = content.replace(doc_full_cmd, new_command)

                with open(str(md_file), "w", encoding="utf-8") as qs_file:
                    qs_file.write(content)
                print(f"Fixed an IPFS hash in doc file {md_file}")
                old_to_new_hashes[doc_hash] = expected_hash
            else:
                print(
                    f"IPFS hash mismatch in doc file {md_file}.\n"
                    f"\tCommand string: {doc_full_cmd}\n"
                    f"\tExpected: {expected_hash}\n"
                    f"\tFound: {doc_hash}\n"
                )

        # Fix package mappings in docs
        for match in [
            m.groupdict() for m in re.finditer(PACKAGE_MAPPING_REGEX, content)
        ]:
            matches += 1
            package_mapping = match["package_mapping"]
            package_hash = match["hash"]

            if package_hash in HASH_SKIPS:
                continue

            expected_hash = package_manager.get_hash_by_attributes(
                match["package_type"], match["vendor"], match["package"]
            )

            if package_hash == expected_hash:
                continue

            hash_mismatches = True

            if fix:
                new_package_mapping = package_mapping.replace(
                    package_hash, expected_hash
                )
                content = content.replace(package_mapping, new_package_mapping)

                with open(str(md_file), "w", encoding="utf-8") as qs_file:
                    qs_file.write(content)
                print(f"Fixed an IPFS hash in doc file {md_file}")
                old_to_new_hashes[package_hash] = expected_hash
            else:
                print(
                    f"IPFS hash mismatch in doc file {md_file}.\n"
                    f"\tMapping string: {package_mapping}\n"
                    f"\tExpected: {expected_hash}\n"
                    f"\tFound: {package_hash}\n"
                )

    # Fix packages in python files
    all_py_files = [Path("autonomy", "constants.py")]
    for py_file in all_py_files:
        content = read_file(str(py_file))
        for match in [m.groupdict() for m in re.finditer(FULL_PACKAGE_REGEX, content)]:
            full_package = match["full_package"]
            py_hash = match["hash"]
            expected_hash = package_manager.get_hash_by_package_line(
                full_package, str(py_file)
            )
            if not expected_hash:
                errors = True
                continue
            expected_package = package_manager.get_package_by_hash(expected_hash)
            if not expected_package:
                errors = True
                continue

            new_package = (":").join(full_package.split(":")[:-1] + [expected_hash])

            # Overwrite with new hash
            if py_hash == expected_hash:
                continue

            hash_mismatches = True

            if fix:
                content = content.replace(full_package, new_package)

                with open(str(py_file), "w", encoding="utf-8") as qs_file:
                    qs_file.write(content)
                print(f"Fixed an IPFS hash in doc file {py_file}")
                old_to_new_hashes[py_hash] = expected_hash
            else:
                print(
                    f"IPFS hash mismatch on file {py_file}.\n"
                    f"\tPackage: {full_package}\n"
                    f"\tExpected: {expected_hash}\n,"
                    f"\tFound: {py_hash}:\n"
                )

    # Fix hashes in package list
    package_list_file = Path(ROOT_DIR, "docs", "package_list.md")
    content = read_file(str(package_list_file))

    for match in [m.groupdict() for m in re.finditer(PACKAGE_TABLE_REGEX, content)]:
        expected_hash = package_manager.get_hash_by_attributes(
            match["package_type"], match["vendor"], match["package"]
        )
        package_hash = match["hash"]

        if package_hash == expected_hash:
            continue

        print(
            f"IPFS hash mismatch in doc file {package_list_file}.\n"
            f'\tPackage: {match["package_type"]}/{match["vendor"]}/{match["package"]}\n'
            f"\tExpected: {expected_hash},\n"
            f"\tFound: {package_hash}\n"
        )
        hash_mismatches = True

        # Overwrite with new hash
        if fix:
            content = content.replace(package_hash, expected_hash)

    if fix:
        with open(str(package_list_file), "w", encoding="utf-8") as p_file:
            p_file.write(content)
        print(f"Fixed some IPFS hashes in doc file {package_list_file}")

    if fix and errors:
        raise ValueError(
            "There were some errors while fixing IPFS hashes. Check the logs."
        )

    if not fix and (hash_mismatches or errors):
        print("There are mismatching IPFS hashes in the docs.")
        sys.exit(1)

    if matches == 0:
        print(
            "No commands were found in the docs. The command regex is probably outdated."
        )
        sys.exit(1)

    print("Checking doc IPFS hashes finished successfully.")


if __name__ == "__main__":
    print("Start checking doc IPFS hashes.")
    parser = argparse.ArgumentParser()
    parser.add_argument("--fix", action="store_true")
    parser.add_argument("-p", "--paths", type=Path, nargs="*", default=[Path("docs")])
    args = parser.parse_args()
    check_ipfs_hashes(paths=args.paths, fix=args.fix)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""On-chain interaction helpers."""

from pathlib import Path
from typing import List, Optional, Tuple, cast

import click
from aea.configurations.data_types import PackageType
from aea.configurations.loader import load_configuration_object
from aea.crypto.base import Crypto, LedgerApi
from aea.crypto.registries import crypto_registry, ledger_apis_registry
from aea.helpers.base import IPFSHash
from texttable import Texttable

from autonomy.chain.base import ServiceState, UnitType
from autonomy.chain.config import ChainConfigs, ChainType, ContractConfigs
from autonomy.chain.constants import (
    AGENT_REGISTRY_CONTRACT,
    COMPONENT_REGISTRY_CONTRACT,
)
from autonomy.chain.exceptions import (
    ComponentMintFailed,
    DependencyError,
    FailedToRetrieveComponentMetadata,
    FailedToRetrieveTokenId,
    InstanceRegistrationFailed,
    InvalidMintParameter,
    ServiceDeployFailed,
    ServiceRegistrationFailed,
    TerminateServiceFailed,
    UnbondServiceFailed,
)
from autonomy.chain.metadata import NFTHashOrPath, publish_metadata
from autonomy.chain.mint import DEFAULT_NFT_IMAGE_HASH
from autonomy.chain.mint import mint_component as _mint_component
from autonomy.chain.mint import mint_service as _mint_service
from autonomy.chain.service import activate_service as _activate_service
from autonomy.chain.service import deploy_service as _deploy_service
from autonomy.chain.service import get_agent_instances, get_service_info
from autonomy.chain.service import register_instance as _register_instance
from autonomy.chain.service import terminate_service as _terminate_service
from autonomy.chain.service import unbond_service as _unbond_service
from autonomy.chain.utils import (
    verify_component_dependencies,
    verify_service_dependencies,
)
from autonomy.configurations.base import PACKAGE_TYPE_TO_CONFIG_CLASS, Service


try:
    from aea_ledger_ethereum.ethereum import EthereumApi, EthereumCrypto

    ETHEREUM_PLUGIN_INSTALLED = True
except ImportError:  # pragma: nocover
    ETHEREUM_PLUGIN_INSTALLED = False

try:
    from aea_ledger_ethereum_hwi.exceptions import HWIError
    from aea_ledger_ethereum_hwi.hwi import EthereumHWIApi

    HWI_PLUGIN_INSTALLED = True
except ImportError:  # pragma: nocover
    HWI_PLUGIN_INSTALLED = False


def get_ledger_and_crypto_objects(
    chain_type: ChainType,
    key: Optional[Path] = None,
    password: Optional[str] = None,
    hwi: bool = False,
) -> Tuple[LedgerApi, Crypto]:
    """Create ledger_api and crypto objects"""

    chain_config = ChainConfigs.get(chain_type=chain_type)
    identifier = EthereumApi.identifier

    if chain_config.rpc is None:
        raise click.ClickException(
            f"RPC URL cannot be `None`, "
            f"Please set the environment variable for {chain_type.value} chain "
            f"using `{ChainConfigs.get_rpc_env_var(chain_type)}` environment variable"
        )

    if hwi and not HWI_PLUGIN_INSTALLED:
        raise click.ClickException(
            "Hardware wallet plugin not installed, "
            "Run `pip3 install open-aea-ledger-ethereum-hwi` to install the plugin"
        )

    if hwi:
        identifier = EthereumHWIApi.identifier

    if not hwi and not ETHEREUM_PLUGIN_INSTALLED:
        raise click.ClickException(
            "Ethereum ledger plugin not installed, "
            "Run `pip3 install open-aea-ledger-ethereum` to install the plugin"
        )

    if key is None:
        crypto = crypto_registry.make(identifier)
    else:
        crypto = EthereumCrypto(
            private_key_path=key,
            password=password,
        )

    ledger_api = ledger_apis_registry.make(
        identifier,
        **{
            "address": chain_config.rpc,
            "chain_id": chain_config.chain_id,
            "is_gas_estimation_enabled": True,
        },
    )

    if hwi:
        # Setting the `LedgerApi.identifier` to `ethereum` for both ledger and
        # hardware plugin to interact with the contract. If we use `ethereum_hwi`
        # as the ledger identifier the contracts will need ABI configuration for
        # the `ethereum_hwi` identifier which means we will have to define hardware
        # wallet as the dependency for contract but the hardware wallet plugin
        # is meant to be used for CLI tools only so we set the identifier to
        # `ethereum` for both ledger and hardware wallet plugin
        ledger_api.identifier = EthereumApi.identifier

    try:
        ledger_api.api.eth.default_account = crypto.address
    except HWIError as e:
        raise click.ClickException(e.message)

    return ledger_api, crypto


def mint_component(  # pylint: disable=too-many-arguments, too-many-locals
    package_path: Path,
    package_type: PackageType,
    key: Optional[Path],
    chain_type: ChainType,
    dependencies: List[int],
    nft: Optional[NFTHashOrPath] = None,
    owner: Optional[str] = None,
    password: Optional[str] = None,
    skip_hash_check: bool = False,
    skip_dependencies_check: bool = False,
    timeout: Optional[float] = None,
    hwi: bool = False,
) -> None:
    """Mint component."""

    is_agent = package_type == PackageType.AGENT
    if is_agent and len(dependencies) == 0:
        raise click.ClickException("Agent packages needs to have dependencies")

    if key is None and not hwi:
        raise click.ClickException(
            "Please provide key path using `--key` or use `--hwi` if you want to use a hardware wallet"
        )

    ledger_api, crypto = get_ledger_and_crypto_objects(
        chain_type=chain_type,
        key=key,
        password=password,
        hwi=hwi,
    )

    try:
        package_configuration = load_configuration_object(
            package_type=package_type,
            directory=package_path,
            package_type_config_class=PACKAGE_TYPE_TO_CONFIG_CLASS,
        )
    except FileNotFoundError as e:  # pragma: nocover
        raise click.ClickException(
            f"Cannot find configuration file for {package_type}"
        ) from e

    if chain_type == ChainType.LOCAL and nft is None:
        nft = IPFSHash(DEFAULT_NFT_IMAGE_HASH)

    if chain_type != ChainType.LOCAL and nft is None:
        raise click.ClickException(
            f"Please provide hash for NFT image to mint component on `{chain_type.value}` chain"
        )

    if not skip_dependencies_check and chain_type == ChainType.ETHEREUM:
        try:
            verify_component_dependencies(
                ledger_api=ledger_api,
                contract_address=ContractConfigs.get(
                    COMPONENT_REGISTRY_CONTRACT.name
                ).contracts[chain_type],
                dependencies=dependencies,
                package_configuration=package_configuration,
                skip_hash_check=skip_hash_check,
            )
        except FailedToRetrieveComponentMetadata as e:
            raise click.ClickException(f"Dependency verification failed; {e}") from e
        except DependencyError as e:
            raise click.ClickException(f"Dependency verification failed; {e}") from e

    metadata_hash, metadata_string = publish_metadata(
        package_id=package_configuration.package_id,
        package_path=package_path,
        nft=cast(str, nft),
        description=package_configuration.description,
    )

    try:
        token_id = _mint_component(
            ledger_api=ledger_api,
            crypto=crypto,
            metadata_hash=metadata_hash,
            owner=owner,
            component_type=UnitType.AGENT if is_agent else UnitType.COMPONENT,
            chain_type=chain_type,
            dependencies=dependencies,
            timeout=timeout,
        )
    except InvalidMintParameter as e:
        raise click.ClickException(f"Invalid parameters provided; {e}") from e
    except ComponentMintFailed as e:
        raise click.ClickException(
            f"Component mint failed with following error; {e}"
        ) from e
    except FailedToRetrieveTokenId as e:
        raise click.ClickException(
            f"Component mint was successful but token ID retrieving failed with following error; {e}"
        ) from e

    click.echo("Component minted with:")
    click.echo(f"\tPublic ID: {package_configuration.public_id}")
    click.echo(f"\tMetadata Hash: {metadata_hash}")
    if token_id is not None:
        click.echo(f"\tToken ID: {token_id}")
        (Path.cwd() / f"{token_id}.json").write_text(metadata_string)
    else:
        raise click.ClickException(
            "Could not verify metadata hash to retrieve the token ID"
        )


def mint_service(  # pylint: disable=too-many-arguments, too-many-locals
    package_path: Path,
    key: Optional[Path],
    chain_type: ChainType,
    agent_id: int,
    number_of_slots: int,
    cost_of_bond: int,
    threshold: int,
    nft: Optional[NFTHashOrPath] = None,
    owner: Optional[str] = None,
    password: Optional[str] = None,
    skip_hash_check: bool = False,
    skip_dependencies_check: bool = False,
    timeout: Optional[float] = None,
    hwi: bool = False,
) -> None:
    """Mint service"""

    if key is None and not hwi:  # pragma: nocover
        raise click.ClickException(
            "Please provide key path using `--key` or use `--hwi` if you want to use a hardware wallet"
        )
    ledger_api, crypto = get_ledger_and_crypto_objects(
        chain_type=chain_type,
        key=key,
        password=password,
        hwi=hwi,
    )

    try:
        package_configuration = load_configuration_object(
            package_type=PackageType.SERVICE,
            directory=package_path,
            package_type_config_class=PACKAGE_TYPE_TO_CONFIG_CLASS,
        )
    except FileNotFoundError as e:  # pragma: nocover
        raise click.ClickException(
            f"Cannot find configuration file for {PackageType.SERVICE}"
        ) from e

    if chain_type == ChainType.LOCAL and nft is None:
        nft = IPFSHash(DEFAULT_NFT_IMAGE_HASH)

    if chain_type != ChainType.LOCAL and nft is None:
        raise click.ClickException(
            f"Please provide hash for NFT image to mint component on `{chain_type.value}` chain"
        )

    if not skip_dependencies_check and chain_type == ChainType.ETHEREUM:
        try:
            verify_service_dependencies(
                ledger_api=ledger_api,
                contract_address=ContractConfigs.get(
                    AGENT_REGISTRY_CONTRACT.name
                ).contracts[chain_type],
                agent_id=agent_id,
                service_configuration=cast(Service, package_configuration),
                skip_hash_check=skip_hash_check,
            )
        except FailedToRetrieveComponentMetadata as e:
            raise click.ClickException(f"Dependency verification failed; {e}") from e
        except DependencyError as e:
            raise click.ClickException(f"Dependency verification failed; {e}") from e

    metadata_hash, metadata_string = publish_metadata(
        package_id=package_configuration.package_id,
        package_path=package_path,
        nft=cast(str, nft),
        description=package_configuration.description,
    )

    try:
        token_id = _mint_service(
            ledger_api=ledger_api,
            crypto=crypto,
            metadata_hash=metadata_hash,
            chain_type=chain_type,
            agent_ids=[
                agent_id,
            ],
            number_of_slots_per_agent=[
                number_of_slots,
            ],
            cost_of_bond_per_agent=[
                cost_of_bond,
            ],
            threshold=threshold,
            timeout=timeout,
            owner=owner,
        )
    except ComponentMintFailed as e:
        raise click.ClickException(
            f"Service mint failed with following error; {e}"
        ) from e
    except FailedToRetrieveTokenId as e:
        raise click.ClickException(
            f"Service mint was successful but token ID retrieving failed with following error; {e}"
        ) from e

    click.echo("Service minted with:")
    click.echo(f"\tPublic ID: {package_configuration.public_id}")
    click.echo(f"\tMetadata Hash: {metadata_hash}")
    if token_id is not None:
        click.echo(f"\tToken ID: {token_id}")
        (Path.cwd() / f"{token_id}.json").write_text(metadata_string)
    else:
        raise click.ClickException(
            "Could not verify metadata hash to retrieve the token ID"
        )


def activate_service(  # pylint: disable=too-many-arguments
    service_id: int,
    key: Path,
    chain_type: ChainType,
    password: Optional[str] = None,
    timeout: Optional[float] = None,
    hwi: bool = False,
) -> None:
    """Activate on-chain service"""

    if key is None and not hwi:  # pragma: nocover
        raise click.ClickException(
            "Please provide key path using `--key` or use `--hwi` if you want to use a hardware wallet"
        )

    ledger_api, crypto = get_ledger_and_crypto_objects(
        chain_type=chain_type,
        key=key,
        password=password,
        hwi=hwi,
    )

    try:
        _activate_service(
            ledger_api=ledger_api,
            crypto=crypto,
            chain_type=chain_type,
            service_id=service_id,
            timeout=timeout,
        )
    except ServiceRegistrationFailed as e:
        raise click.ClickException(str(e)) from e

    click.echo("Service activated succesfully")


def register_instance(  # pylint: disable=too-many-arguments
    service_id: int,
    instances: List[str],
    agent_ids: List[int],
    key: Path,
    chain_type: ChainType,
    password: Optional[str] = None,
    timeout: Optional[float] = None,
    hwi: bool = False,
) -> None:
    """Register agents instances on an activated service"""

    if key is None and not hwi:  # pragma: nocover
        raise click.ClickException(
            "Please provide key path using `--key` or use `--hwi` if you want to use a hardware wallet"
        )

    ledger_api, crypto = get_ledger_and_crypto_objects(
        chain_type=chain_type,
        key=key,
        password=password,
        hwi=hwi,
    )

    try:
        _register_instance(
            ledger_api=ledger_api,
            crypto=crypto,
            chain_type=chain_type,
            service_id=service_id,
            instances=instances,
            agent_ids=agent_ids,
            timeout=timeout,
        )
    except InstanceRegistrationFailed as e:
        raise click.ClickException(str(e)) from e

    click.echo("Agent instance registered succesfully")


def deploy_service(  # pylint: disable=too-many-arguments
    service_id: int,
    key: Path,
    chain_type: ChainType,
    deployment_payload: Optional[str] = None,
    password: Optional[str] = None,
    timeout: Optional[float] = None,
    hwi: bool = False,
) -> None:
    """Deploy a service with registration activated"""

    if key is None and not hwi:  # pragma: nocover
        raise click.ClickException(
            "Please provide key path using `--key` or use `--hwi` if you want to use a hardware wallet"
        )

    ledger_api, crypto = get_ledger_and_crypto_objects(
        chain_type=chain_type,
        key=key,
        password=password,
        hwi=hwi,
    )

    try:
        _deploy_service(
            ledger_api=ledger_api,
            crypto=crypto,
            chain_type=chain_type,
            service_id=service_id,
            deployment_payload=deployment_payload,
            timeout=timeout,
        )
    except ServiceDeployFailed as e:
        raise click.ClickException(str(e)) from e

    click.echo("Service deployed succesfully")


def terminate_service(
    service_id: int,
    key: Path,
    chain_type: ChainType,
    password: Optional[str] = None,
    hwi: bool = False,
) -> None:
    """Terminate a service"""

    if key is None and not hwi:  # pragma: nocover
        raise click.ClickException(
            "Please provide key path using `--key` or use `--hwi` if you want to use a hardware wallet"
        )

    ledger_api, crypto = get_ledger_and_crypto_objects(
        chain_type=chain_type,
        key=key,
        password=password,
        hwi=hwi,
    )

    try:
        _terminate_service(
            ledger_api=ledger_api,
            crypto=crypto,
            chain_type=chain_type,
            service_id=service_id,
        )
    except TerminateServiceFailed as e:
        raise click.ClickException(str(e)) from e

    click.echo("Service terminated succesfully")


def unbond_service(
    service_id: int,
    key: Path,
    chain_type: ChainType,
    password: Optional[str] = None,
    hwi: bool = False,
) -> None:
    """Terminate a service"""

    if key is None and not hwi:  # pragma: nocover
        raise click.ClickException(
            "Please provide key path using `--key` or use `--hwi` if you want to use a hardware wallet"
        )

    ledger_api, crypto = get_ledger_and_crypto_objects(
        chain_type=chain_type,
        key=key,
        password=password,
        hwi=hwi,
    )

    try:
        _unbond_service(
            ledger_api=ledger_api,
            crypto=crypto,
            chain_type=chain_type,
            service_id=service_id,
        )
    except UnbondServiceFailed as e:
        raise click.ClickException(str(e)) from e

    click.echo("Service unbonded succesfully")


def print_service_info(
    service_id: int,
    chain_type: ChainType,
) -> None:
    """Terminate a service"""

    ledger_api, _ = get_ledger_and_crypto_objects(
        chain_type=chain_type,
    )
    (
        security_deposit,
        multisig_address,
        _,
        threshold,
        max_agents,
        number_of_agent_instances,
        _service_state,
        cannonical_agents,
    ) = get_service_info(
        ledger_api=ledger_api,
        chain_type=chain_type,
        token_id=service_id,
    )
    service_state = ServiceState(_service_state)
    rows = [
        ("Property", "Value"),
        ("Service State", service_state.name),
        ("Security Deposit", security_deposit),
        ("Multisig Address", multisig_address),
        ("Cannonical Agents", ", ".join(map(str, cannonical_agents))),
        ("Max Agents", max_agents),
        ("Threshold", threshold),
        ("Number Of Agent Instances", number_of_agent_instances),
    ]

    if service_state.value >= ServiceState.ACTIVE_REGISTRATION.value:
        rows.append(
            (
                "Registered Instances",
                "\n".join(
                    map(
                        lambda x: f"- {x}",
                        get_agent_instances(
                            ledger_api=ledger_api,
                            chain_type=chain_type,
                            token_id=service_id,
                        ).get("agentInstances", []),
                    ),
                ),
            ),
        )
    click.echo(Texttable().add_rows(rows=rows).draw())



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tools to build and run agents from existing deployments."""

import contextlib
import os
import shutil
import signal
import subprocess  # nosec
import sys
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, List, Optional

from autonomy.deploy.base import TENDERMINT_COM_URL_PARAM, TENDERMINT_URL_PARAM


ETHEREUM_PRIVATE_KEY_FILE = "ethereum_private_key.txt"

CONNECTION_ABCI_CONFIG_HOST = "CONNECTION_ABCI_CONFIG_HOST"
CONNECTION_ABCI_CONFIG_PORT = "CONNECTION_ABCI_CONFIG_PORT"


class AgentRunner:
    """Agent runner."""

    agent_id: int
    agent_data: Dict[str, str]
    registry_path: Path
    process: Optional[subprocess.Popen]  # nosec
    aea_cli: List[str] = [sys.executable, "-m", "aea.cli"]
    agent_alias: str = "agent"

    def __init__(self, agent_id: int, agent_data: Dict, registry_path: Path) -> None:
        """Initialize object."""

        self.agent_id = agent_id
        self.agent_data = agent_data
        self.registry_path = registry_path
        self.agent_env = os.environ.copy()
        self.agent_dir = TemporaryDirectory()  # pylint: disable=consider-using-with
        self.cwd = Path(".").resolve().absolute()

        agent_env_data = self.agent_data["environment"]
        for env_var in agent_env_data:
            key, value = env_var.split("=")
            if key.endswith(TENDERMINT_URL_PARAM.upper()) or key.endswith(
                TENDERMINT_COM_URL_PARAM.upper()
            ):
                self.agent_env[key] = f"http://localhost:8080/{self.agent_id}"
            else:
                self.agent_env[key] = value

        self.agent_env[CONNECTION_ABCI_CONFIG_HOST] = "localhost"
        self.agent_env[CONNECTION_ABCI_CONFIG_PORT] = f"2665{self.agent_id}"

    def start(
        self,
    ) -> None:
        """Start process."""

        agent_dir = Path(self.agent_dir.name)
        os.chdir(agent_dir)

        # TODO: replace with more appropriate env var name; we mean an agent here
        print(f"Loading {self.agent_env['VALORY_APPLICATION']}")
        subprocess.run(  # nosec  # pylint: disable=subprocess-run-check
            [
                *self.aea_cli,
                "--registry-path",
                str(self.registry_path),
                "fetch",
                self.agent_env["VALORY_APPLICATION"],
                "--local",
                "--alias",
                self.agent_alias,
            ],
            env=self.agent_env,
        )
        os.chdir(agent_dir / self.agent_alias)
        Path(agent_dir, self.agent_alias, ETHEREUM_PRIVATE_KEY_FILE).write_text(
            self.agent_env["AEA_KEY"], encoding="utf-8"
        )

        subprocess.run(  # nosec # pylint: disable=subprocess-run-check
            [*self.aea_cli, "add-key", "ethereum"], env=self.agent_env
        )
        subprocess.run(  # nosec # pylint: disable=subprocess-run-check
            ["cp", "ethereum_private_key.txt", "ethereum_flashbots_private_key.txt"],
            env=self.agent_env,
        )
        subprocess.run(  # nosec # pylint: disable=subprocess-run-check
            [*self.aea_cli, "add-key", "ethereum-flashbots"], env=self.agent_env
        )
        subprocess.run(  # nosec # pylint: disable=subprocess-run-check
            [*self.aea_cli, "install"], env=self.agent_env
        )
        self.process = subprocess.Popen(  # nosec # pylint: disable=subprocess-run-check,consider-using-with
            [*self.aea_cli, "run", "--aev"], env=self.agent_env
        )

    def stop(
        self,
    ) -> None:
        """Stop the process."""

        os.chdir(str(self.cwd))
        with contextlib.suppress(OSError, PermissionError, FileNotFoundError):
            shutil.rmtree(str(self.agent_dir))

        if self.process is None:
            return  # pragma: no cover

        self.process.poll()
        if self.process.returncode is None:  # stop only pending processes
            os.kill(self.process.pid, signal.SIGTERM)

        self.process.wait(timeout=5)
        self.process.terminate()
        self.process = None



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Implementation of the 'autonomy fetch' subcommand."""

from typing import cast

import click
from aea.cli.fetch import NotAnAgentPackage, do_fetch
from aea.cli.utils.click_utils import PublicIdParameter, registry_flag
from aea.cli.utils.context import Context
from aea.configurations.base import PublicId
from aea.configurations.constants import AGENT, SERVICE

from autonomy.cli.helpers.registry import fetch_service


@click.command(name="fetch")
@registry_flag()
@click.option(
    "--alias",
    type=str,
    required=False,
    help="Provide a local alias for the agent.",
)
@click.option(
    "--agent",
    "package_type",
    help="Specify the package type as agent (default).",
    default=True,
    flag_value=AGENT,
)
@click.option(
    "--service",
    "package_type",
    help="Specify the package type as service.",
    flag_value=SERVICE,
)
@click.argument("public-id", type=PublicIdParameter(), required=True)
@click.pass_context
def fetch(
    click_context: click.Context,
    public_id: PublicId,
    alias: str,
    package_type: str,
    registry: str,
) -> None:
    """Fetch an agent from the registry."""
    ctx = cast(Context, click_context.obj)
    ctx.registry_type = registry

    try:
        if package_type == AGENT:
            do_fetch(ctx, public_id, alias)
        else:
            fetch_service(ctx, public_id, alias)
    except NotAnAgentPackage as e:
        raise click.ClickException(
            "Downloaded packages is not an agent package, "
            "if you intend to download a service please use `--service` flag "
            "or check the hash"
        ) from e
    except Exception as e:  # pylint: disable=broad-except
        raise click.ClickException(str(e)) from e



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Override for hash command."""

import sys
from pathlib import Path
from typing import Optional
from warnings import warn

import click
from aea.cli.ipfs_hash import hash_file, to_v0_string, to_v1_string

from autonomy.cli.helpers.ipfs_hash import load_configuration, update_hashes


@click.group(name="hash")
def hash_group() -> None:
    """Hashing utils."""


@hash_group.command(name="all")
@click.option(
    "--packages-dir",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    default=Path("packages/"),
)
@click.option("--vendor", type=str)
@click.option("--no-wrap", is_flag=True)
def generate_all(
    packages_dir: Path,
    vendor: Optional[str],
    no_wrap: bool,
) -> None:
    """Generate IPFS hashes."""
    message = (
        "`autonomy hash all` command has been deprecated and will be removed on v2.0.0, "
        "please use `autonomy packages lock` command to perform package hash updates"
    )
    warn(
        message=message,
        category=DeprecationWarning,
        stacklevel=2,
    )
    click.echo(message=message)
    packages_dir = Path(packages_dir).absolute()
    return_code = update_hashes(
        packages_dir, no_wrap, vendor=vendor, config_loader=load_configuration
    )
    sys.exit(return_code)


hash_group.add_command(hash_file)
hash_group.add_command(to_v0_string)
hash_group.add_command(to_v1_string)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains round behaviours of Background Behaviours."""

import logging
import platform
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Type
from unittest import mock

import pytest

from packages.valory.contracts.gnosis_safe.contract import GnosisSafeContract
from packages.valory.contracts.multisend.contract import MultiSendContract
from packages.valory.contracts.service_registry.contract import ServiceRegistryContract
from packages.valory.protocols.contract_api import ContractApiMessage
from packages.valory.protocols.contract_api.custom_types import RawTransaction, State
from packages.valory.skills.abstract_round_abci.base import AbciAppDB
from packages.valory.skills.abstract_round_abci.behaviour_utils import AsyncBehaviour
from packages.valory.skills.abstract_round_abci.behaviours import BaseBehaviour
from packages.valory.skills.abstract_round_abci.test_tools.base import (
    FSMBehaviourBaseCase,
)
from packages.valory.skills.termination_abci import PUBLIC_ID
from packages.valory.skills.termination_abci.behaviours import (
    BackgroundBehaviour,
    TerminationAbciBehaviours,
    TerminationBehaviour,
)
from packages.valory.skills.termination_abci.rounds import Event, SynchronizedData
from packages.valory.skills.transaction_settlement_abci.behaviours import (
    TransactionSettlementRoundBehaviour,
)


SERVICE_REGISTRY_ADDRESS = "0x48b6af7B12C71f09e2fC8aF4855De4Ff54e775cA"
SAFE_ADDRESS = "0x0"
MULTISEND_ADDRESS = "0xA238CBeb142c10Ef7Ad8442C6D1f9E89e07e7761"
SERVICE_OWNER_ADDRESS = "0x0"
SERVICE_ID = None


def test_skill_public_id() -> None:
    """Test skill module public ID"""

    # pylint: disable=no-member
    assert PUBLIC_ID.name == Path(__file__).parents[1].name
    assert PUBLIC_ID.author == Path(__file__).parents[3].name


@dataclass
class BehaviourTestCase:
    """BehaviourTestCase"""

    name: str
    initial_data: Dict[str, Any]
    ok_reqs: List[Callable]
    err_reqs: List[Callable]
    expected_log: str
    expected_log_level: int


class BaseTerminationTest(FSMBehaviourBaseCase):
    """Base test case."""

    path_to_skill = Path(__file__).parents[1]

    behaviour: TerminationAbciBehaviours
    behaviour_class: Type[BaseBehaviour]
    next_behaviour_class: Type[
        BaseBehaviour
    ] = TransactionSettlementRoundBehaviour.initial_behaviour_cls
    synchronized_data: SynchronizedData
    done_event = Event.TERMINATE

    def fast_forward(self, data: Optional[Dict[str, Any]] = None) -> None:
        """Fast-forward on initialization"""

        data = data or {}
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.behaviour_class.auto_behaviour_id(),
            SynchronizedData(AbciAppDB(setup_data=AbciAppDB.data_to_lists(data))),
        )
        assert (
            self.behaviour.current_behaviour is not None
            and self.behaviour.current_behaviour.behaviour_id
            == self.behaviour_class.auto_behaviour_id()
        )

    def complete(self) -> None:
        """Complete test"""
        self.behaviour.act_wrapper()
        self.mock_a2a_transaction()
        self.end_round(done_event=self.done_event)
        assert (
            self.behaviour.current_behaviour is not None
            and self.behaviour.current_behaviour.behaviour_id
            == self.next_behaviour_class.auto_behaviour_id()
        )


class TestBackgroundBehaviour(BaseTerminationTest):
    """Test the background behaviour."""

    behaviour_class = BackgroundBehaviour
    next_behaviour_class = TransactionSettlementRoundBehaviour.initial_behaviour_cls

    # when the service owner ordered termination
    _ZERO_TRANSFER_EVENTS = [
        {"block_number": 1},
        {"block_number": 10},
        {"block_number": 20},
    ]
    _NUM_ZERO_TRANSFER_EVENTS = len(_ZERO_TRANSFER_EVENTS)
    # when the service owner got removed from being a safe owner
    # this happens when the operators take over the safe again
    _SERVICE_OWNER_REMOVED_EVENTS = [
        {"block_number": 2},
        {"block_number": 11},
    ]
    _NUM_SERVICE_OWNER_REMOVED_EVENTS = len(_SERVICE_OWNER_REMOVED_EVENTS)
    _SAFE_OWNERS = ["0x1", "0x2", "0x3", "0x4"]
    _NUM_SAFE_OWNERS = len(_SAFE_OWNERS)
    _SAFE_THRESHOLD = 1
    _MOCK_TX_RESPONSE = b"0xIrrelevantForTests".hex()
    _MOCK_TX_HASH = "0x" + "0" * 64
    _INITIAL_DATA: Dict[str, Any] = dict(
        all_participants=_SAFE_OWNERS,
        safe_contract_address=SAFE_ADDRESS,
        participants=_SAFE_OWNERS,
        consensus_threshold=3,
    )

    _STATE_ERR_LOG = (
        f"Expected response performative {ContractApiMessage.Performative.STATE.value}, "  # type: ignore
        f"received {ContractApiMessage.Performative.ERROR}."
    )
    _RAW_TRANSACTION_ERR = (
        f"Expected response performative {ContractApiMessage.Performative.RAW_TRANSACTION.value}, "  # type: ignore
        f"received {ContractApiMessage.Performative.ERROR}."
    )
    _SERVICE_OWNER_ERR_LOG = (
        f"Couldn't get the service owner for service with id={SERVICE_ID}. "
        f"{_STATE_ERR_LOG}"
    )
    _ZERO_TRANSFER_EVENTS_ERR_LOG = (
        f"Couldn't get the latest Zero Transfer (`SafeReceived`) event. "
        f"{_STATE_ERR_LOG}"
    )
    _SERVICE_OWNER_REMOVED_EVENTS_ERR_LOG = (
        f"Couldn't get the latest `RemovedOwner` event. " f"{_STATE_ERR_LOG}"
    )
    _SAFE_OWNERS_ERR_LOG = (
        f"Couldn't get the safe owners for safe deployed at {SAFE_ADDRESS}. "
        f"{_STATE_ERR_LOG}"
    )
    _REMOVE_OWNER_ERR_LOG = f"Couldn't get a remove owner tx. " f"{_STATE_ERR_LOG}"
    _SWAP_OWNER_ERR_LOG = f"Couldn't get a swap owner tx. " f"{_STATE_ERR_LOG}"
    _SAFE_HASH_ERR_LOG = f"Couldn't get safe hash. " f"{_STATE_ERR_LOG}"
    _MULTISEND_ERR_LOG = "Couldn't compile the multisend tx. "
    _SUCCESS_LOG = "Successfully prepared termination multisend tx."
    _IS_STOPPED_LOG = "dropping message as behaviour has stopped:"
    _IS_NOT_WAITING_MESSAGE = "could not send message"

    def _mock_get_service_owner_request(
        self,
        error: bool = False,
    ) -> None:
        """Mock a ServiceRegistryContract.get_service_owner() request."""
        if not error:
            response_performative = ContractApiMessage.Performative.STATE
            response_body = dict(service_owner=SERVICE_OWNER_ADDRESS)
        else:
            response_body = dict()
            response_performative = ContractApiMessage.Performative.ERROR

        self.mock_contract_api_request(
            contract_id=str(ServiceRegistryContract.contract_id),
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_STATE,
                contract_address=SERVICE_REGISTRY_ADDRESS,
            ),
            response_kwargs=dict(
                performative=response_performative,
                state=State(
                    ledger_id="ethereum",
                    body=response_body,
                ),
            ),
        )

    def _mock_get_zero_transfer_events_request(
        self,
        error: bool = False,
        num_events: int = _NUM_ZERO_TRANSFER_EVENTS,
    ) -> None:
        """Mock a GnosisSafeContract.get_zero_transfer_events() request."""
        if not error:
            response_performative = ContractApiMessage.Performative.STATE
            response_body = dict(data=self._ZERO_TRANSFER_EVENTS[:num_events])
        else:
            response_body = dict()
            response_performative = ContractApiMessage.Performative.ERROR
        self.mock_contract_api_request(
            contract_id=str(GnosisSafeContract.contract_id),
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_STATE,
                contract_address=SAFE_ADDRESS,
            ),
            response_kwargs=dict(
                performative=response_performative,
                state=State(
                    ledger_id="ethereum",
                    body=response_body,
                ),
            ),
        )

    def _mock_get_removed_owner_events_request(
        self,
        error: bool = False,
        num_events: int = _NUM_SERVICE_OWNER_REMOVED_EVENTS,
    ) -> None:
        """Mock a GnosisSafeContract.get_removed_owner_events() request."""
        if not error:
            response_performative = ContractApiMessage.Performative.STATE
            response_body = dict(data=self._SERVICE_OWNER_REMOVED_EVENTS[:num_events])
        else:
            response_body = dict()
            response_performative = ContractApiMessage.Performative.ERROR

        self.mock_contract_api_request(
            contract_id=str(GnosisSafeContract.contract_id),
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_STATE,
                contract_address=SAFE_ADDRESS,
            ),
            response_kwargs=dict(
                performative=response_performative,
                state=State(
                    ledger_id="ethereum",
                    body=response_body,
                ),
            ),
        )

    def _mock_get_owners_request(
        self,
        error: bool = False,
    ) -> None:
        """Mock a GnosisSafeContract.get_owners() request."""
        if not error:
            response_performative = ContractApiMessage.Performative.STATE
            response_body = dict(owners=self._SAFE_OWNERS)
        else:
            response_body = dict()
            response_performative = ContractApiMessage.Performative.ERROR
        self.mock_contract_api_request(
            contract_id=str(GnosisSafeContract.contract_id),
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_STATE,
                contract_address=SAFE_ADDRESS,
            ),
            response_kwargs=dict(
                performative=response_performative,
                state=State(
                    ledger_id="ethereum",
                    body=response_body,
                ),
            ),
        )

    def _mock_get_remove_owner_data_request(self, error: bool = False) -> None:
        """Mock a GnosisSafeContract.get_remove_owner_data() request."""
        if not error:
            response_performative = ContractApiMessage.Performative.STATE
            response_body = dict(data=self._MOCK_TX_RESPONSE)
        else:
            response_body = dict()
            response_performative = ContractApiMessage.Performative.ERROR
        self.mock_contract_api_request(
            contract_id=str(GnosisSafeContract.contract_id),
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_STATE,
                contract_address=SAFE_ADDRESS,
            ),
            response_kwargs=dict(
                performative=response_performative,
                state=State(
                    ledger_id="ethereum",
                    body=response_body,
                ),
            ),
        )

    def _mock_get_swap_owner_data_request(
        self,
        error: bool = False,
    ) -> None:
        """Mock a GnosisSafeContract.get_swap_owner_data() request."""
        if not error:
            response_performative = ContractApiMessage.Performative.STATE
            response_body = dict(data=self._MOCK_TX_RESPONSE)
        else:
            response_body = dict()
            response_performative = ContractApiMessage.Performative.ERROR
        self.mock_contract_api_request(
            contract_id=str(GnosisSafeContract.contract_id),
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_STATE,
                contract_address=SAFE_ADDRESS,
            ),
            response_kwargs=dict(
                performative=response_performative,
                state=State(
                    ledger_id="ethereum",
                    body=response_body,
                ),
            ),
        )

    def _mock_get_raw_safe_transaction_hash_request(
        self,
        error: bool = False,
    ) -> None:
        """Mock a GnosisSafeContract.get_raw_safe_transaction_hash() request."""
        if not error:
            response_performative = ContractApiMessage.Performative.STATE
            response_body = dict(tx_hash=self._MOCK_TX_HASH)
        else:
            response_body = dict()
            response_performative = ContractApiMessage.Performative.ERROR
        self.mock_contract_api_request(
            contract_id=str(GnosisSafeContract.contract_id),
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_STATE,
                contract_address=SAFE_ADDRESS,
            ),
            response_kwargs=dict(
                performative=response_performative,
                state=State(
                    ledger_id="ethereum",
                    body=response_body,
                ),
            ),
        )

    def _mock_get_tx_data_request(
        self,
        error: bool = False,
    ) -> None:
        """Mock a MultiSendContract.get_tx_data() request."""
        if not error:
            response_performative = ContractApiMessage.Performative.RAW_TRANSACTION
            response_body = dict(data=self._MOCK_TX_RESPONSE)
        else:
            response_body = dict()
            response_performative = ContractApiMessage.Performative.ERROR
        self.mock_contract_api_request(
            contract_id=str(MultiSendContract.contract_id),
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
                contract_address=MULTISEND_ADDRESS,
            ),
            response_kwargs=dict(
                performative=response_performative,
                raw_transaction=RawTransaction(
                    ledger_id="ethereum",
                    body=response_body,
                ),
            ),
        )

    def _mock_is_stopped(  # pylint: disable=unused-argument, disable=protected-access
        self,
        error: bool = False,
    ) -> None:
        """Mock a MultiSendContract.get_tx_data() request."""
        assert self.behaviour.current_behaviour is not None
        self.behaviour.current_behaviour._AsyncBehaviour__stopped = True

    def _mock_state_is_not_waiting_message(  # pylint: disable=unused-argument, disable=protected-access
        self,
        error: bool = False,
    ) -> None:
        """Mock a MultiSendContract.get_tx_data() request."""
        assert self.behaviour.current_behaviour is not None
        self.behaviour.current_behaviour._AsyncBehaviour__state = (
            AsyncBehaviour.AsyncState.RUNNING
        )

    @pytest.mark.parametrize(
        "test_case",
        [
            BehaviourTestCase(
                name="agent fails to get the service owner",
                initial_data=_INITIAL_DATA,
                ok_reqs=[],
                err_reqs=[_mock_get_service_owner_request],
                expected_log=_SERVICE_OWNER_ERR_LOG,
                expected_log_level=logging.ERROR,
            ),
            BehaviourTestCase(
                name="agent fails to get zero transfer event",
                initial_data=_INITIAL_DATA,
                ok_reqs=[_mock_get_service_owner_request],
                err_reqs=[_mock_get_zero_transfer_events_request],
                expected_log=_ZERO_TRANSFER_EVENTS_ERR_LOG,
                expected_log_level=logging.ERROR,
            ),
            BehaviourTestCase(
                name="agent fails to get owner removal event",
                initial_data=_INITIAL_DATA,
                ok_reqs=[
                    _mock_get_service_owner_request,
                    _mock_get_zero_transfer_events_request,
                ],
                err_reqs=[_mock_get_removed_owner_events_request],
                expected_log=_SERVICE_OWNER_REMOVED_EVENTS_ERR_LOG,
                expected_log_level=logging.ERROR,
            ),
            BehaviourTestCase(
                name="agent fails to get the safe owners",
                initial_data=_INITIAL_DATA,
                ok_reqs=[
                    _mock_get_service_owner_request,
                    _mock_get_zero_transfer_events_request,
                    _mock_get_removed_owner_events_request,
                ],
                err_reqs=[_mock_get_owners_request],
                expected_log=_SAFE_OWNERS_ERR_LOG,
                expected_log_level=logging.ERROR,
            ),
            BehaviourTestCase(
                name="agent fails to get a remove safe owner tx",
                initial_data=_INITIAL_DATA,
                ok_reqs=[
                    _mock_get_service_owner_request,
                    _mock_get_zero_transfer_events_request,
                    _mock_get_removed_owner_events_request,
                    _mock_get_owners_request,
                ],
                err_reqs=[_mock_get_remove_owner_data_request],
                expected_log=_REMOVE_OWNER_ERR_LOG,
                expected_log_level=logging.ERROR,
            ),
            BehaviourTestCase(
                name="agent fails to get a swap safe owner tx",
                initial_data=_INITIAL_DATA,
                ok_reqs=[
                    _mock_get_service_owner_request,
                    _mock_get_zero_transfer_events_request,
                    _mock_get_removed_owner_events_request,
                    _mock_get_owners_request,
                    *[_mock_get_remove_owner_data_request] * (_NUM_SAFE_OWNERS - 1),
                ],
                err_reqs=[_mock_get_swap_owner_data_request],
                expected_log=_SWAP_OWNER_ERR_LOG,
                expected_log_level=logging.ERROR,
            ),
            BehaviourTestCase(
                name="agent fails to prepare multisend tx",
                initial_data=_INITIAL_DATA,
                ok_reqs=[
                    _mock_get_service_owner_request,
                    _mock_get_zero_transfer_events_request,
                    _mock_get_removed_owner_events_request,
                    _mock_get_owners_request,
                    *[_mock_get_remove_owner_data_request] * (_NUM_SAFE_OWNERS - 1),
                    _mock_get_swap_owner_data_request,
                ],
                err_reqs=[_mock_get_tx_data_request],
                expected_log=_MULTISEND_ERR_LOG,
                expected_log_level=logging.ERROR,
            ),
            BehaviourTestCase(
                name="agent fails to get a hash for the multisend tx",
                initial_data=_INITIAL_DATA,
                ok_reqs=[
                    _mock_get_service_owner_request,
                    _mock_get_zero_transfer_events_request,
                    _mock_get_removed_owner_events_request,
                    _mock_get_owners_request,
                    *[_mock_get_remove_owner_data_request] * (_NUM_SAFE_OWNERS - 1),
                    _mock_get_swap_owner_data_request,
                    _mock_get_tx_data_request,
                ],
                err_reqs=[_mock_get_raw_safe_transaction_hash_request],
                expected_log=_SAFE_HASH_ERR_LOG,
                expected_log_level=logging.ERROR,
            ),
            BehaviourTestCase(
                name="agent completes the whole flow",
                initial_data=_INITIAL_DATA,
                ok_reqs=[
                    _mock_get_service_owner_request,
                    _mock_get_zero_transfer_events_request,
                    _mock_get_removed_owner_events_request,
                    _mock_get_owners_request,
                    *[_mock_get_remove_owner_data_request] * (_NUM_SAFE_OWNERS - 1),
                    _mock_get_swap_owner_data_request,
                    _mock_get_tx_data_request,
                    _mock_get_raw_safe_transaction_hash_request,
                ],
                err_reqs=[],
                expected_log=_SUCCESS_LOG,
                expected_log_level=logging.INFO,
            ),
            BehaviourTestCase(
                name="agent drops message because app already stopped",
                initial_data=_INITIAL_DATA,
                ok_reqs=[],
                err_reqs=[
                    _mock_is_stopped,
                    _mock_get_service_owner_request,
                ],
                expected_log=_IS_STOPPED_LOG,
                expected_log_level=logging.INFO,
            ),
            BehaviourTestCase(
                name="agent could not send message because state != WAITING_MESSAGE",
                initial_data=_INITIAL_DATA,
                ok_reqs=[],
                err_reqs=[
                    _mock_state_is_not_waiting_message,
                    _mock_get_service_owner_request,
                ],
                expected_log=_IS_NOT_WAITING_MESSAGE,
                expected_log_level=logging.WARNING,
            ),
        ],
    )
    def test_run(self, test_case: BehaviourTestCase) -> None:
        """Test multiple paths of termination."""
        self.fast_forward(data=test_case.initial_data)
        # repeating this check for the `current_behaviour` here to avoid `mypy` reporting:
        # `error: Item "None" of "Optional[BaseBehaviour]" has no attribute "context"` when accessing the context below
        assert self.behaviour.current_behaviour is not None

        with mock.patch.object(
            self.behaviour.current_behaviour.context.logger, "log"
        ) as mock_logger:
            self.behaviour.act_wrapper()

            # apply the OK mocks first
            for ok_req in test_case.ok_reqs:
                ok_req(self)

            # apply the failing mocks
            for err_req in test_case.err_reqs:
                err_req(self, error=True)

            log_found = False
            for log_args in mock_logger.call_args_list:
                if platform.python_version().startswith("3.7"):
                    actual_log_level, actual_log = log_args[0][:2]
                else:
                    actual_log_level, actual_log = log_args.args[:2]

                if actual_log.startswith(test_case.expected_log):
                    assert actual_log_level == test_case.expected_log_level, (
                        f"{test_case.expected_log} was expected to log on {test_case.expected_log_level} log level, "
                        f"but logged on {log_args[0]} instead."
                    )
                    log_found = True
                    break

            if not log_found:
                raise AssertionError(
                    f'Expected log message "{test_case.expected_log}" was not found in captured logs: '
                    f"{mock_logger.call_args_list}."
                )

        if len(test_case.err_reqs) == 0:
            # no mocked requests fail,
            # the behaviour should complete
            self.complete()

    def test_termination_majority_already_reached(self) -> None:
        """Tests the background behaviour when the termination is already reached."""
        self.fast_forward(
            data=dict(
                termination_majority_reached=True,
                all_participants=["a"],
                participants=["a"],
                consensus_threshold=1,
            )
        )
        with mock.patch.object(
            self.behaviour_class, "check_for_signal"
        ) as check_for_signal:
            self.behaviour.act_wrapper()
            check_for_signal.assert_not_called()

    def test_no_termination_signal_is_present(self) -> None:
        """Tests the background behaviour when no termination signal is present."""
        self.fast_forward(self._INITIAL_DATA)
        with mock.patch.object(AsyncBehaviour, "sleep") as sleep:
            self.behaviour.act_wrapper()
            self._mock_get_service_owner_request()
            self._mock_get_zero_transfer_events_request(num_events=0)
            sleep.assert_called()

    def test_no_remove_owner_event_is_present(self) -> None:
        """Tests the background behaviour when the safe owner hasn't been removed."""
        self.fast_forward(self._INITIAL_DATA)
        self.behaviour.act_wrapper()
        self._mock_get_service_owner_request()
        self._mock_get_zero_transfer_events_request()
        self._mock_get_removed_owner_events_request(num_events=0)
        self._mock_get_owners_request()
        for _ in range(self._NUM_SAFE_OWNERS - 1):
            self._mock_get_remove_owner_data_request()
        self._mock_get_swap_owner_data_request()
        self._mock_get_tx_data_request()
        self._mock_get_raw_safe_transaction_hash_request()
        self.complete()


class TestTerminationBehaviour(BaseTerminationTest):
    """Test termination behaviour."""

    behaviour_class = TerminationBehaviour

    def test_run(self) -> None:
        """Test that the Termination Behaviour exits the app."""
        self.fast_forward()
        with mock.patch("sys.exit") as sys_exit:
            self.behaviour.act_wrapper()
            sys_exit.assert_called()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains tests for the documentation tools."""

import re

from scripts.check_doc_ipfs_hashes import AEA_COMMAND_REGEX, FULL_PACKAGE_REGEX


def test_cmd_regex() -> None:
    """Test the command regex"""

    lines = [
        "aea fetch --remote open_aea/my_first_aea:bafybeibnjfr3sdg57ggyxbcfkh42yqkj6a3gftp55l26aaw2z2jvvc3tny",
        "aea fetch open_aea/my_first_aea:bafybeibnjfr3sdg57ggyxbcfkh42yqkj6a3gftp55l26aaw2z2jvvc3tny",
        "aea fetch --remote --other_flag open_aea/my_first_aea:bafybeibnjfr3sdg57ggyxbcfkh42yqkj6a3gftp55l26aaw2z2jvvc3tny",
        "autonomy fetch --remote open_aea/my_first_aea:bafybeibnjfr3sdg57ggyxbcfkh42yqkj6a3gftp55l26aaw2z2jvvc3tny",
        "autonomy fetch open_aea/my_first_aea:bafybeibnjfr3sdg57ggyxbcfkh42yqkj6a3gftp55l26aaw2z2jvvc3tny",
        "autonomy fetch --remote --other_flag open_aea/my_first_aea:bafybeibnjfr3sdg57ggyxbcfkh42yqkj6a3gftp55l26aaw2z2jvvc3tny",
        "autonomy fetch valory/hello_world:0.1.0:bafybeihh6fz5xti3vmzd4ktihcvhqknnijmnc5vgbbqjcanqlfimqis6zy --remote --service",
    ]

    for line in lines:
        assert re.match(
            AEA_COMMAND_REGEX, line
        ), f"line '{line}' does not match AEA_COMMAND_REGEX"


def test_package_regex() -> None:
    """Test the command regex"""

    lines = [
        "valory/abstract_round_abci:0.1.0:bafybeibmwz5b7khd2efbegr7djknjdtwfrm6ufmsyiyhoxwz6r6p5cljv4",
        "open_aea/signing:bafybeid6f5ceool4evgaxdegs5pkjoivqjk3mo3ehctalswngb5m5c32wm",
    ]

    for line in lines:
        assert re.match(
            FULL_PACKAGE_REGEX, line
        ), f"line '{line}' does not match FULL_PACKAGE_REGEX"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test for cli click utils."""

import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Optional, Tuple, Type

import click
import pytest
from aea.helpers.base import IPFSHash
from aea.test_tools.click_testing import CliRunner

from autonomy.analyse.abci.app_spec import FSMSpecificationLoader as FSMSpecLoader
from autonomy.chain.config import ChainType
from autonomy.chain.mint import DEFAULT_NFT_IMAGE_HASH
from autonomy.cli.utils.click_utils import (
    NFTArgument,
    PathArgument,
    abci_spec_format_flag,
    chain_selection_flag,
    image_profile_flag,
    sys_path_patch,
)
from autonomy.deploy.image import ImageProfiles

from tests.conftest import DATA_DIR


@dataclass
class ClickFlagTestCase:
    """ClickFlagTestCase"""

    decorator: Callable
    items: Tuple
    opt_prefix: str = ""


@pytest.mark.parametrize(
    "test_case",
    [
        ClickFlagTestCase(image_profile_flag, ImageProfiles.ALL),
        ClickFlagTestCase(
            chain_selection_flag,
            tuple(map(lambda x: x.value, ChainType)),
            opt_prefix="use-",
        ),
        ClickFlagTestCase(abci_spec_format_flag, FSMSpecLoader.OutputFormats.ALL),
    ],
)
def test_flag_decorators(test_case: ClickFlagTestCase) -> None:
    """Test click flag decorators"""

    @click.command()
    @test_case.decorator()
    def dummy() -> None:
        """Dummy"""

    assert len(dummy.params) == len(test_case.items)
    for parameter, item in zip(dummy.params, reversed(test_case.items)):
        assert isinstance(parameter, click.Option)
        assert item in parameter.flag_value
        assert parameter.opts == [f"--{test_case.opt_prefix}" + item.replace("_", "-")]
        assert parameter.help


def test_path_argument() -> None:
    """Test PathArgument"""

    @click.command()
    @click.option("--path", type=PathArgument())
    def dummy(path: Optional[Path]) -> None:
        """Dummy"""
        click.echo(isinstance(path, Path))

    cli_runner = CliRunner()
    result = cli_runner.invoke(dummy, args=("--path", "packages"))
    assert result.stdout.strip() == str(True)


def test_sys_path_patch() -> None:
    """Test sys_path_patch"""

    cwd = "dummy"
    assert not cwd == sys.path[0]
    with sys_path_patch(Path(cwd)):
        assert cwd == sys.path[0]
    assert not cwd == sys.path[0]


@pytest.mark.parametrize(
    ("value", "itype"),
    (
        (DEFAULT_NFT_IMAGE_HASH, IPFSHash),
        (str(DATA_DIR / "nft.png"), Path),
    ),
)
def test_nft_argument(value: str, itype: Type) -> None:
    """Test `NFTArgument` flag"""

    assert isinstance(NFTArgument().convert(value, None, None), itype)


def test_nft_argument_failure() -> None:
    """Test `NFTArgument` flag failure"""

    with pytest.raises(
        click.ClickException,
        match="Invalid value provided for the NFT image argument, please provide a valid IPFS hash or valid image path",
    ):
        NFTArgument().convert(str(DATA_DIR / "not_nft.png"), None, None)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Implementation of the `autonomy service` command"""

from pathlib import Path
from typing import List, Optional

import click
from aea.cli.utils.context import Context
from aea.cli.utils.decorators import pass_ctx

from autonomy.chain.config import ChainType
from autonomy.cli.helpers.chain import (
    activate_service,
    deploy_service,
    print_service_info,
    register_instance,
    terminate_service,
    unbond_service,
)
from autonomy.cli.mint import (
    hwi_flag,
    key_path_decorator,
    password_decorator,
    timeout_flag,
)
from autonomy.cli.utils.click_utils import chain_selection_flag


@click.group("service")
@pass_ctx
@chain_selection_flag()
@timeout_flag
def service(ctx: Context, chain_type: str, timeout: float) -> None:
    """Manage on-chain services."""

    ctx.config["chain_type"] = ChainType(chain_type)
    ctx.config["timeout"] = timeout


@service.command()
@pass_ctx
@click.argument("service_id", type=int)
@key_path_decorator
@hwi_flag
@password_decorator
def activate(
    ctx: Context,
    service_id: int,
    key: Path,
    hwi: bool,
    password: Optional[str],
) -> None:
    """Activate service."""

    activate_service(
        service_id=service_id,
        key=key,
        hwi=hwi,
        chain_type=ctx.config["chain_type"],
        password=password,
        timeout=ctx.config["timeout"],
    )


@service.command()
@pass_ctx
@click.argument("service_id", type=int)
@click.option(
    "-i",
    "--instance",
    "instances",
    type=str,
    required=True,
    multiple=True,
    help="Agent instance address",
)
@click.option(
    "-a",
    "--agent-id",
    "agent_ids",
    type=int,
    required=True,
    multiple=True,
    help="Agent ID",
)
@key_path_decorator
@hwi_flag
@password_decorator
def register(  # pylint: disable=too-many-arguments
    ctx: Context,
    service_id: int,
    instances: List[str],
    agent_ids: List[int],
    key: Path,
    hwi: bool,
    password: Optional[str],
) -> None:
    """Register instances."""

    register_instance(
        service_id=service_id,
        instances=instances,
        agent_ids=agent_ids,
        key=key,
        hwi=hwi,
        chain_type=ctx.config["chain_type"],
        password=password,
        timeout=ctx.config["timeout"],
    )


@service.command()
@pass_ctx
@click.argument("service_id", type=int)
@click.option(
    "-d",
    "--deployment-payload",
    type=int,
    help="Deployment payload value",
)
@key_path_decorator
@hwi_flag
@password_decorator
def deploy(
    ctx: Context,
    service_id: int,
    key: Path,
    hwi: bool,
    password: Optional[str],
    deployment_payload: Optional[str],
) -> None:
    """Deploy a service."""

    deploy_service(
        service_id=service_id,
        key=key,
        hwi=hwi,
        chain_type=ctx.config["chain_type"],
        password=password,
        deployment_payload=deployment_payload,
    )


@service.command(name="terminate")
@pass_ctx
@click.argument("service_id", type=int)
@key_path_decorator
@hwi_flag
@password_decorator
def _terminate(
    ctx: Context,
    service_id: int,
    key: Path,
    hwi: bool,
    password: Optional[str],
) -> None:
    """Terminate a service."""

    terminate_service(
        service_id=service_id,
        key=key,
        hwi=hwi,
        chain_type=ctx.config["chain_type"],
        password=password,
    )


@service.command(name="unbond")
@pass_ctx
@click.argument("service_id", type=int)
@key_path_decorator
@hwi_flag
@password_decorator
def _unbond(
    ctx: Context,
    service_id: int,
    key: Path,
    hwi: bool,
    password: Optional[str],
) -> None:
    """Unbond a service."""

    unbond_service(
        service_id=service_id,
        key=key,
        hwi=hwi,
        chain_type=ctx.config["chain_type"],
        password=password,
    )


@service.command(name="info")
@pass_ctx
@click.argument("service_id", type=int)
def _info(
    ctx: Context,
    service_id: int,
) -> None:
    """Print service information."""

    print_service_info(
        service_id=service_id,
        chain_type=ctx.config["chain_type"],
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Usefule click utils."""
import contextlib
import copy
import sys
from pathlib import Path
from typing import Any, Callable, Generator, Optional, Union, cast

import click
from aea.cli.utils.config import get_default_author_from_cli_config
from aea.helpers.base import IPFSHash, SimpleId

from autonomy.analyse.abci.app_spec import FSMSpecificationLoader
from autonomy.chain.config import ChainType
from autonomy.deploy.image import ImageProfiles


def image_profile_flag(
    default: str = ImageProfiles.PRODUCTION, mark_default: bool = True
) -> Callable:
    """Choice of one flag between: '--local/--remote'."""

    def wrapper(f: Callable) -> Callable:
        for profile in ImageProfiles.ALL:
            f = click.option(
                f"--{profile}",
                "profile",
                flag_value=profile,
                help=f"To use the {profile} profile.",
                default=(profile == default) and mark_default,
            )(f)

        return f

    return wrapper


def abci_spec_format_flag(
    default: str = FSMSpecificationLoader.OutputFormats.YAML, mark_default: bool = True
) -> Callable:
    """Flags for abci spec outputs formats."""

    def wrapper(f: Callable) -> Callable:
        for of in FSMSpecificationLoader.OutputFormats.ALL:
            f = click.option(
                f"--{of}",
                "spec_format",
                flag_value=of,
                help=f"{of.title()} file.",
                default=(of == default) and mark_default,
            )(f)

        return f

    return wrapper


def chain_selection_flag(
    default: ChainType = ChainType.LOCAL,
    mark_default: bool = True,
    help_string_format: str = "To use {} chain profile to interact with the contracts",
) -> Callable:
    """Flags for abci spec outputs formats."""

    def wrapper(f: Callable) -> Callable:
        for chain_type in ChainType:
            chain_name = cast(str, chain_type.value).replace("_", "-")
            f = click.option(
                f"--use-{chain_name}",
                "chain_type",
                flag_value=chain_type.value,
                help=help_string_format.format(chain_name),
                default=(chain_type == default) and mark_default,
            )(f)
        return f

    return wrapper


@contextlib.contextmanager
def sys_path_patch(path: Path) -> Generator:
    """Patch sys.path variable with new import path at highest priority."""
    old_sys_path = copy.copy(sys.path)
    sys.path.insert(0, str(path))
    yield
    sys.path = old_sys_path


class PathArgument(click.Path):
    """Path parameter for CLI."""

    def convert(
        self, value: Any, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> Optional[Path]:
        """Convert path string to `pathlib.Path`"""
        path_string = super().convert(value, param, ctx)
        return None if path_string is None else Path(path_string)


class NFTArgument(click.ParamType):
    """NFT parameter for minting tools."""

    METAVAR = "IPFS_HASH_OR_IMAGE_PATH"

    def get_metavar(self, param: click.Parameter) -> str:  # pragma: nocover
        """Get metavar"""
        return self.METAVAR

    def convert(
        self, value: Any, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> Optional[Union[Path, IPFSHash]]:
        """Convert path string to `pathlib.Path`"""
        if value is None:  # pragma: nocover
            return None
        try:
            return IPFSHash(value)
        except ValueError as e:
            image_path = Path(value)
            if not (image_path.exists() and image_path.is_file()):
                raise click.ClickException(
                    "Invalid value provided for the NFT image argument, "
                    "please provide a valid IPFS hash or valid image path"
                ) from e
            return image_path


def image_author_option(fn: Callable) -> Callable:
    """Wrap function with clik option for image-author"""

    def _validate(_ctx, _param, value):  # type: ignore
        if value is None:
            return get_default_author_from_cli_config()
        SimpleId(value)
        return value

    return click.option(
        "--image-author",
        type=str,
        help="Specify author name for docker image.",
        default=None,
        callback=_validate,
    )(fn)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Fuzzy tests for valory/abci connection"""
import os

from hypothesis import settings

from packages.valory.connections.abci import CI
from packages.valory.connections.abci.tests.test_fuzz.base import BaseFuzzyTests
from packages.valory.connections.abci.tests.test_fuzz.mock_node.channels.grpc_channel import (
    GrpcChannel,
)
from packages.valory.connections.abci.tests.test_fuzz.mock_node.channels.tcp_channel import (
    TcpChannel,
)


running_on_ci = os.getenv(CI)
if running_on_ci:
    settings.load_profile(CI)


class TestFuzzyGrpc(BaseFuzzyTests):
    """Test the connection when gRPC is used"""

    CHANNEL_TYPE = GrpcChannel
    USE_GRPC = True
    AGENT_TIMEOUT_SECONDS = 30


class TestFuzzyTcp(BaseFuzzyTests):
    """Test the connection when TCP is used"""

    CHANNEL_TYPE = TcpChannel
    USE_GRPC = False
    AGENT_TIMEOUT_SECONDS = 30



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Database schemas and helpers"""

import sqlite3
from datetime import datetime
from pathlib import Path
from typing import Any, Iterator, List, Optional, Tuple

from autonomy.analyse.logs.base import LogRow


TIMESTAMP = "timestamp"
LOG_LEVEL = "log_level"
MESSAGE = "message"
PERIOD = "period"
ROUND = "round_name"
BEHAVIOUR = "behaviour_name"
EXIT_EVENT = "exit_event"

QUERY_CREATE_LOG_TABLE = (
    "CREATE TABLE {agent} "
    + f"({TIMESTAMP} TIMESTAMP, {LOG_LEVEL} TEXT, {MESSAGE} TEXT, {PERIOD} INTEGER, {ROUND} TEXT, {BEHAVIOUR} TEXT);"
)
QUERY_CHECK_TABLE_EXISTS = (
    "SELECT name FROM sqlite_master WHERE type='table' AND name=?;"
)
QUERY_DROP_TABLE = "DROP TABLE {agent};"
QUERY_INSERT_LOG = "INSERT INTO {agent} VALUES (?, ?, ?, ?, ?, ?);"


class AgentLogsDB:
    """Logs DB"""

    _db: sqlite3.Connection

    def __init__(self, agent: str, file: Path) -> None:
        """Initialize object."""

        self.agent = agent
        self._db_path = file
        self._db = sqlite3.connect(
            database=self._db_path,
            detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES,
        )

    def select(  # pylint: disable=too-many-arguments
        self,
        start_time: Optional[datetime] = None,
        end_time: Optional[datetime] = None,
        log_level: Optional[str] = None,
        period: Optional[int] = None,
        round_name: Optional[str] = None,
        behaviour_name: Optional[str] = None,
    ) -> List[LogRow]:
        """Build select query."""

        query = f"SELECT * from {self.agent}"  # nosec
        paramaters: List[Any] = []

        def _append_condition(_query: str, condition: str) -> str:
            """Append condition at the end of the query."""
            if "WHERE" in query:
                return query + " AND " + condition
            return query + " WHERE " + condition

        if start_time is not None:
            query = _append_condition(_query=query, condition=f"{TIMESTAMP}>?")
            paramaters.append(start_time)

        if end_time is not None:
            query = _append_condition(_query=query, condition=f"{TIMESTAMP}<?")
            paramaters.append(end_time)

        if log_level is not None:
            query = _append_condition(_query=query, condition=f"{LOG_LEVEL}=?")
            paramaters.append(log_level)

        if period is not None:
            query = _append_condition(_query=query, condition=f"{PERIOD}=?")
            paramaters.append(period)

        if round_name is not None:
            query = _append_condition(_query=query, condition=f"{ROUND}=?")
            paramaters.append(round_name)

        if behaviour_name is not None:
            query = _append_condition(_query=query, condition=f"{BEHAVIOUR}=?")
            paramaters.append(behaviour_name)

        query += ";"
        return self.cursor.execute(query, paramaters).fetchall()

    def execution_path(self) -> List[Tuple[int, str, str]]:
        """Extraction FSM execution path"""
        return self.cursor.execute(
            f"SELECT {PERIOD}, {ROUND}, MAX({EXIT_EVENT}) from {self.agent} GROUP BY {ROUND};"
        ).fetchall()

    @property
    def cursor(self) -> sqlite3.Cursor:
        """Creates and returns a database cursor."""
        return self._db.cursor()

    def exists(self) -> bool:
        """Check if table already exists."""

        return (
            self.cursor.execute(QUERY_CHECK_TABLE_EXISTS, (self.agent,)).fetchone()
            is not None
        )

    def delete(self) -> "AgentLogsDB":
        """Delete table"""
        self.cursor.execute(QUERY_DROP_TABLE.format(agent=self.agent))
        return self

    def create(self, reset: bool = False) -> "AgentLogsDB":
        """Create agent table"""

        exists = self.exists()
        if exists and not reset:
            return self

        if exists and reset:
            self.delete()

        self.cursor.execute(QUERY_CREATE_LOG_TABLE.format(agent=self.agent))
        self._db.commit()

        return self

    def insert_many(
        self,
        logs: Iterator[LogRow],
    ) -> "AgentLogsDB":
        """Insert a record"""
        for timestamp, log_level, message, period, round_name, behaviour_name in logs:
            self.cursor.execute(
                QUERY_INSERT_LOG.format(agent=self.agent),
                (timestamp, log_level, message, period, round_name, behaviour_name),
            )
        self._db.commit()
        return self



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Script to create environment for benchmarking n agents."""

from pathlib import Path
from typing import Any, Dict, List, Optional, cast

import yaml
from aea.configurations.constants import DEFAULT_LEDGER

from autonomy.constants import (
    HARDHAT_IMAGE_NAME,
    HARDHAT_IMAGE_VERSION,
    OAR_IMAGE,
    TENDERMINT_IMAGE_NAME,
    TENDERMINT_IMAGE_VERSION,
)
from autonomy.deploy.base import BaseDeploymentGenerator, ServiceBuilder
from autonomy.deploy.constants import (
    DEFAULT_ENCODING,
    KEY_SCHEMA_PRIVATE_KEY,
    KEY_SCHEMA_TYPE,
    KUBERNETES_AGENT_KEY_NAME,
)
from autonomy.deploy.generators.kubernetes.templates import (
    AGENT_NODE_TEMPLATE,
    AGENT_SECRET_TEMPLATE,
    CLUSTER_CONFIGURATION_TEMPLATE,
    HARDHAT_TEMPLATE,
    PORTS_CONFIG_DEPLOYMENT,
    PORT_CONFIG_DEPLOYMENT,
)


class KubernetesGenerator(BaseDeploymentGenerator):
    """Kubernetes Deployment Generator."""

    output_name: str = "build.yaml"
    deployment_type: str = "kubernetes"

    def __init__(  # pylint: disable=too-many-arguments
        self,
        service_builder: ServiceBuilder,
        build_dir: Path,
        dev_mode: bool = False,
        packages_dir: Optional[Path] = None,
        open_aea_dir: Optional[Path] = None,
        open_autonomy_dir: Optional[Path] = None,
        use_tm_testnet_setup: bool = False,
        image_author: Optional[str] = None,
    ) -> None:
        """Initialise the deployment generator."""
        super().__init__(
            service_builder=service_builder,
            build_dir=build_dir,
            use_tm_testnet_setup=use_tm_testnet_setup,
            dev_mode=dev_mode,
            packages_dir=packages_dir,
            open_aea_dir=open_aea_dir,
            open_autonomy_dir=open_autonomy_dir,
            image_author=image_author,
        )
        self.resources: List[str] = []

    def build_agent_deployment(
        self,
        runtime_image: str,
        agent_ix: int,
        number_of_agents: int,
        agent_vars: Dict[str, Any],
        agent_ports: Optional[Dict[int, int]] = None,
    ) -> str:
        """Build agent deployment."""

        host_names = ", ".join(
            [f'"--hostname=abci{i}"' for i in range(number_of_agents)]
        )

        agent_ports_deployment = ""
        if agent_ports is not None:
            port_mappings = map(
                lambda x: PORT_CONFIG_DEPLOYMENT.format(port=x[0]),
                agent_ports.items(),
            )
            agent_ports_deployment = "\n".join(
                [PORTS_CONFIG_DEPLOYMENT, *port_mappings]
            )

        agent_deployment = AGENT_NODE_TEMPLATE.format(
            runtime_image=runtime_image,
            validator_ix=agent_ix,
            aea_key=self.service_builder.keys[agent_ix][KEY_SCHEMA_PRIVATE_KEY],
            number_of_validators=number_of_agents,
            host_names=host_names,
            tendermint_image_name=TENDERMINT_IMAGE_NAME,
            tendermint_image_version=TENDERMINT_IMAGE_VERSION,
            log_level=self.service_builder.log_level,
            agent_ports_deployment=agent_ports_deployment,
            ledger=self.service_builder.keys[agent_ix].get(
                KEY_SCHEMA_TYPE, DEFAULT_LEDGER
            ),
        )
        agent_deployment_yaml = yaml.load_all(agent_deployment, Loader=yaml.FullLoader)  # type: ignore
        resources = []
        for resource in agent_deployment_yaml:
            if resource.get("kind") == "Deployment":
                for container in resource["spec"]["template"]["spec"]["containers"]:
                    if container["name"] == "aea":
                        container["env"] += [
                            {"name": k, "value": f"{v}"} for k, v in agent_vars.items()
                        ]
            resources.append(resource)

        res = "\n---\n".join([yaml.safe_dump(i) for i in resources])
        return res

    def generate_config_tendermint(self) -> "KubernetesGenerator":
        """Build configuration job."""

        if self.tendermint_job_config is not None:  # pragma: no cover
            return self

        host_names = ", ".join(
            [
                f'"--hostname=abci{i}"'
                for i in range(self.service_builder.service.number_of_agents)
            ]
        )

        self.tendermint_job_config = CLUSTER_CONFIGURATION_TEMPLATE.format(
            valory_app=self.service_builder.service.agent.name,
            number_of_validators=self.service_builder.service.number_of_agents,
            host_names=host_names,
            tendermint_image_name=TENDERMINT_IMAGE_NAME,
            tendermint_image_version=TENDERMINT_IMAGE_VERSION,
        )

        return self

    def generate(
        self,
        image_version: Optional[str] = None,
        use_hardhat: bool = False,
        use_acn: bool = False,
    ) -> "KubernetesGenerator":
        """Generate the deployment."""

        image_version = image_version or self.service_builder.service.agent.hash

        if self.dev_mode:
            image_version = "dev"

        if use_hardhat:
            self.resources.append(
                HARDHAT_TEMPLATE % (HARDHAT_IMAGE_NAME, HARDHAT_IMAGE_VERSION)
            )

        if use_acn:
            # insert an ACN node into resources
            ...

        agent_vars = self.service_builder.generate_agents()  # type:ignore
        runtime_image = OAR_IMAGE.format(
            image_author=self.image_author,
            agent=self.service_builder.service.agent.name,
            version=image_version,
        )

        agents = "\n---\n".join(
            [
                self.build_agent_deployment(
                    runtime_image=runtime_image,
                    agent_ix=i,
                    number_of_agents=self.service_builder.service.number_of_agents,
                    agent_vars=agent_vars[i],
                    agent_ports=(
                        self.service_builder.service.deployment_config.get("agent", {})
                        .get("ports", {})
                        .get(i)
                    ),
                )
                for i in range(self.service_builder.service.number_of_agents)
            ]
        )
        self.resources.append(agents)
        self.output = "\n---\n".join(self.resources)
        return self

    def write_config(
        self,
    ) -> "KubernetesGenerator":
        """Write output to build dir"""

        output = "---\n".join([self.output, cast(str, self.tendermint_job_config)])
        if not self.build_dir.is_dir():  # pragma: no cover
            self.build_dir.mkdir()
        with open(
            self.build_dir / self.output_name, "w", encoding=DEFAULT_ENCODING
        ) as f:
            f.write(output)

        return self

    def populate_private_keys(self) -> "BaseDeploymentGenerator":
        """Populates private keys into a config map for the kubernetes deployment."""
        path = self.build_dir / "agent_keys"
        for x in range(self.service_builder.service.number_of_agents):
            ledger = self.service_builder.keys[x].get(KEY_SCHEMA_TYPE, DEFAULT_LEDGER)
            key = self.service_builder.keys[x][KEY_SCHEMA_PRIVATE_KEY]
            secret = AGENT_SECRET_TEMPLATE.format(
                private_key=key, validator_ix=x, ledger=ledger
            )
            with open(
                path / KUBERNETES_AGENT_KEY_NAME.format(agent_n=x), "w", encoding="utf8"
            ) as f:
                f.write(secret)
        return self



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/abstract_round_abci skill's behaviours."""
import json
from abc import ABC
from copy import copy
from enum import Enum
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, Type, cast
from unittest import mock
from unittest.mock import MagicMock

from aea.helpers.transaction.base import SignedMessage
from aea.test_tools.test_skill import BaseSkillTestCase

from packages.open_aea.protocols.signing import SigningMessage
from packages.valory.connections.http_client.connection import (
    PUBLIC_ID as HTTP_CLIENT_PUBLIC_ID,
)
from packages.valory.connections.ledger.connection import (
    PUBLIC_ID as LEDGER_CONNECTION_PUBLIC_ID,
)
from packages.valory.protocols.contract_api import ContractApiMessage
from packages.valory.protocols.http import HttpMessage
from packages.valory.protocols.ledger_api.message import LedgerApiMessage
from packages.valory.skills.abstract_round_abci.base import (
    AbstractRound,
    BaseSynchronizedData,
    BaseTxPayload,
    OK_CODE,
    _MetaPayload,
)
from packages.valory.skills.abstract_round_abci.behaviours import (
    AbstractRoundBehaviour,
    BaseBehaviour,
)
from packages.valory.skills.abstract_round_abci.handlers import (
    ContractApiHandler,
    HttpHandler,
    LedgerApiHandler,
    SigningHandler,
    TendermintHandler,
)


# pylint: disable=protected-access,too-few-public-methods,consider-using-with


class FSMBehaviourBaseCase(BaseSkillTestCase, ABC):
    """Base case for testing FSMBehaviour classes."""

    path_to_skill: Path
    behaviour: AbstractRoundBehaviour
    ledger_handler: LedgerApiHandler
    http_handler: HttpHandler
    contract_handler: ContractApiHandler
    signing_handler: SigningHandler
    tendermint_handler: TendermintHandler
    old_tx_type_to_payload_cls: Dict[str, Type[BaseTxPayload]]
    benchmark_dir: TemporaryDirectory
    default_ledger: str = "ethereum"

    @classmethod
    def setup_class(cls, **kwargs: Any) -> None:
        """Setup the test class."""
        if not hasattr(cls, "path_to_skill"):
            raise ValueError(f"No `path_to_skill` set on {cls}")  # pragma: nocover
            # works once https://github.com/valory-xyz/open-aea/issues/492 is fixed
        # we need to store the current value of the meta-class attribute
        # _MetaPayload.transaction_type_to_payload_cls, and restore it
        # in the teardown function. We do a shallow copy so we avoid
        # to modify the old mapping during the execution of the tests.
        cls.old_tx_type_to_payload_cls = copy(_MetaPayload.registry)
        _MetaPayload.registry = {}
        super().setup_class(**kwargs)  # pylint: disable=no-value-for-parameter
        assert (
            cls._skill.skill_context._agent_context is not None
        ), "Agent context not set"  # nosec
        cls._skill.skill_context._agent_context.identity._default_address_key = (
            cls.default_ledger
        )
        cls._skill.skill_context._agent_context._default_ledger_id = cls.default_ledger
        behaviour = cls._skill.skill_context.behaviours.main
        assert isinstance(
            behaviour, AbstractRoundBehaviour
        ), f"{behaviour} is not of type {AbstractRoundBehaviour}"
        cls.behaviour = behaviour
        for attr, handler, handler_type in [
            ("http_handler", cls._skill.skill_context.handlers.http, HttpHandler),
            (
                "signing_handler",
                cls._skill.skill_context.handlers.signing,
                SigningHandler,
            ),
            (
                "contract_handler",
                cls._skill.skill_context.handlers.contract_api,
                ContractApiHandler,
            ),
            (
                "ledger_handler",
                cls._skill.skill_context.handlers.ledger_api,
                LedgerApiHandler,
            ),
            (
                "tendermint_handler",
                cls._skill.skill_context.handlers.tendermint,
                TendermintHandler,
            ),
        ]:
            assert isinstance(
                handler, handler_type
            ), f"{handler} is not of type {handler_type}"
            setattr(cls, attr, handler)

        if kwargs.get("param_overrides") is not None:
            for param_name, param_value in kwargs["param_overrides"].items():
                cls.behaviour.context.params.__dict__[param_name] = param_value

    def setup(self, **kwargs: Any) -> None:
        """
        Set up the test method.

        Called each time before a test method is called.

        :param kwargs: the keyword arguments passed to _prepare_skill
        """
        super().setup(**kwargs)
        self.behaviour.setup()
        self._skill.skill_context.state.setup()
        self._skill.skill_context.state.round_sequence.end_sync()

        self.benchmark_dir = TemporaryDirectory()
        self._skill.skill_context.benchmark_tool.__dict__["log_dir"] = Path(
            self.benchmark_dir.name
        )
        assert (  # nosec
            cast(BaseBehaviour, self.behaviour.current_behaviour).behaviour_id
            == self.behaviour.initial_behaviour_cls.auto_behaviour_id()
        )

    def fast_forward_to_behaviour(
        self,
        behaviour: AbstractRoundBehaviour,
        behaviour_id: str,
        synchronized_data: BaseSynchronizedData,
    ) -> None:
        """Fast forward the FSM to a behaviour."""
        next_behaviour = {s.auto_behaviour_id(): s for s in behaviour.behaviours}[
            behaviour_id
        ]
        next_behaviour = cast(Type[BaseBehaviour], next_behaviour)
        behaviour.current_behaviour = next_behaviour(
            name=next_behaviour.auto_behaviour_id(), skill_context=behaviour.context
        )
        self.skill.skill_context.state.round_sequence.abci_app._round_results.append(
            synchronized_data
        )
        self.skill.skill_context.state.round_sequence.abci_app._extend_previous_rounds_with_current_round()
        self.skill.skill_context.behaviours.main._last_round_height = (
            self.skill.skill_context.state.round_sequence.abci_app.current_round_height
        )
        self.skill.skill_context.state.round_sequence.abci_app._current_round_cls = (
            next_behaviour.matching_round
        )
        # consensus parameters will not be available if the current skill is abstract
        consensus_params = getattr(
            self.skill.skill_context.params, "consensus_params", None
        )
        self.skill.skill_context.state.round_sequence.abci_app._current_round = (
            next_behaviour.matching_round(synchronized_data, consensus_params)
        )

    def mock_ledger_api_request(
        self, request_kwargs: Dict, response_kwargs: Dict
    ) -> None:
        """
        Mock http request.

        :param request_kwargs: keyword arguments for request check.
        :param response_kwargs: keyword arguments for mock response.
        """

        self.assert_quantity_in_outbox(1)
        actual_ledger_api_message = self.get_message_from_outbox()
        assert actual_ledger_api_message is not None, "No message in outbox."  # nosec
        has_attributes, error_str = self.message_has_attributes(
            actual_message=actual_ledger_api_message,
            message_type=LedgerApiMessage,
            to=str(LEDGER_CONNECTION_PUBLIC_ID),
            sender=str(self.skill.skill_context.skill_id),
            **request_kwargs,
        )

        assert has_attributes, error_str  # nosec
        incoming_message = self.build_incoming_message(
            message_type=LedgerApiMessage,
            dialogue_reference=(
                actual_ledger_api_message.dialogue_reference[0],
                "stub",
            ),
            target=actual_ledger_api_message.message_id,
            message_id=-1,
            to=str(self.skill.skill_context.skill_id),
            sender=str(LEDGER_CONNECTION_PUBLIC_ID),
            ledger_id=str(LEDGER_CONNECTION_PUBLIC_ID),
            **response_kwargs,
        )
        self.ledger_handler.handle(incoming_message)
        self.behaviour.act_wrapper()

    def mock_contract_api_request(
        self, contract_id: str, request_kwargs: Dict, response_kwargs: Dict
    ) -> None:
        """
        Mock http request.

        :param contract_id: contract id.
        :param request_kwargs: keyword arguments for request check.
        :param response_kwargs: keyword arguments for mock response.
        """

        self.assert_quantity_in_outbox(1)
        actual_contract_ledger_message = self.get_message_from_outbox()
        assert (  # nosec
            actual_contract_ledger_message is not None
        ), "No message in outbox."
        has_attributes, error_str = self.message_has_attributes(
            actual_message=actual_contract_ledger_message,
            message_type=ContractApiMessage,
            to=str(LEDGER_CONNECTION_PUBLIC_ID),
            sender=str(self.skill.skill_context.skill_id),
            ledger_id="ethereum",
            contract_id=contract_id,
            message_id=1,
            **request_kwargs,
        )
        assert has_attributes, error_str  # nosec
        self.behaviour.act_wrapper()

        incoming_message = self.build_incoming_message(
            message_type=ContractApiMessage,
            dialogue_reference=(
                actual_contract_ledger_message.dialogue_reference[0],
                "stub",
            ),
            target=actual_contract_ledger_message.message_id,
            message_id=-1,
            to=str(self.skill.skill_context.skill_id),
            sender=str(LEDGER_CONNECTION_PUBLIC_ID),
            ledger_id="ethereum",
            contract_id="mock_contract_id",
            **response_kwargs,
        )
        self.contract_handler.handle(incoming_message)
        self.behaviour.act_wrapper()

    def mock_http_request(self, request_kwargs: Dict, response_kwargs: Dict) -> None:
        """
        Mock http request.

        :param request_kwargs: keyword arguments for request check.
        :param response_kwargs: keyword arguments for mock response.
        """

        self.assert_quantity_in_outbox(1)
        actual_http_message = self.get_message_from_outbox()
        assert actual_http_message is not None, "No message in outbox."  # nosec
        has_attributes, error_str = self.message_has_attributes(
            actual_message=actual_http_message,
            message_type=HttpMessage,
            performative=HttpMessage.Performative.REQUEST,
            to=str(HTTP_CLIENT_PUBLIC_ID),
            sender=str(self.skill.skill_context.skill_id),
            **request_kwargs,
        )
        assert has_attributes, error_str  # nosec
        self.behaviour.act_wrapper()
        self.assert_quantity_in_outbox(0)
        incoming_message = self.build_incoming_message(
            message_type=HttpMessage,
            dialogue_reference=(actual_http_message.dialogue_reference[0], "stub"),
            performative=HttpMessage.Performative.RESPONSE,
            target=actual_http_message.message_id,
            message_id=-1,
            to=str(self.skill.skill_context.skill_id),
            sender=str(HTTP_CLIENT_PUBLIC_ID),
            **response_kwargs,
        )
        self.http_handler.handle(incoming_message)
        self.behaviour.act_wrapper()

    def mock_signing_request(self, request_kwargs: Dict, response_kwargs: Dict) -> None:
        """Mock signing request."""
        self.assert_quantity_in_decision_making_queue(1)
        actual_signing_message = self.get_message_from_decision_maker_inbox()
        assert actual_signing_message is not None, "No message in outbox."  # nosec
        has_attributes, error_str = self.message_has_attributes(
            actual_message=actual_signing_message,
            message_type=SigningMessage,
            to=self.skill.skill_context.decision_maker_address,
            sender=str(self.skill.skill_context.skill_id),
            **request_kwargs,
        )
        assert has_attributes, error_str  # nosec
        incoming_message = self.build_incoming_message(
            message_type=SigningMessage,
            dialogue_reference=(actual_signing_message.dialogue_reference[0], "stub"),
            target=actual_signing_message.message_id,
            message_id=-1,
            to=str(self.skill.skill_context.skill_id),
            sender=self.skill.skill_context.decision_maker_address,
            **response_kwargs,
        )
        self.signing_handler.handle(incoming_message)
        self.behaviour.act_wrapper()

    def mock_a2a_transaction(
        self,
    ) -> None:
        """Performs mock a2a transaction."""

        self.mock_signing_request(
            request_kwargs=dict(
                performative=SigningMessage.Performative.SIGN_MESSAGE,
            ),
            response_kwargs=dict(
                performative=SigningMessage.Performative.SIGNED_MESSAGE,
                signed_message=SignedMessage(
                    ledger_id="ethereum", body="stub_signature"
                ),
            ),
        )

        self.mock_http_request(
            request_kwargs=dict(
                method="GET",
                headers="",
                version="",
                body=b"",
            ),
            response_kwargs=dict(
                version="",
                status_code=200,
                status_text="",
                headers="",
                body=json.dumps({"result": {"hash": "", "code": OK_CODE}}).encode(
                    "utf-8"
                ),
            ),
        )
        self.mock_http_request(
            request_kwargs=dict(
                method="GET",
                headers="",
                version="",
                body=b"",
            ),
            response_kwargs=dict(
                version="",
                status_code=200,
                status_text="",
                headers="",
                body=json.dumps({"result": {"tx_result": {"code": OK_CODE}}}).encode(
                    "utf-8"
                ),
            ),
        )

    def end_round(self, done_event: Enum) -> None:
        """Ends round early to cover `wait_for_end` generator."""
        current_behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        if current_behaviour is None:
            return
        current_behaviour = cast(BaseBehaviour, current_behaviour)
        abci_app = current_behaviour.context.state.round_sequence.abci_app
        old_round = abci_app._current_round
        abci_app._last_round = old_round
        abci_app._current_round = abci_app.transition_function[
            current_behaviour.matching_round
        ][done_event](abci_app.synchronized_data)
        abci_app._previous_rounds.append(old_round)
        abci_app._current_round_height += 1
        self.behaviour._process_current_round()

    def _test_done_flag_set(self) -> None:
        """Test that, when round ends, the 'done' flag is set."""
        current_behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert not current_behaviour.is_done()  # nosec
        with mock.patch.object(
            self.behaviour.context.state, "_round_sequence"
        ) as mock_round_sequence:
            mock_round_sequence.last_round_id = cast(
                AbstractRound, current_behaviour.matching_round
            ).auto_round_id()
            current_behaviour.act_wrapper()
            assert current_behaviour.is_done()  # nosec

    @classmethod
    def teardown_class(cls) -> None:
        """Teardown the test class."""
        if getattr(cls, "old_tx_type_to_payload_cls", False):
            _MetaPayload.registry = cls.old_tx_type_to_payload_cls

    def teardown(self, **kwargs: Any) -> None:
        """Teardown."""
        super().teardown(**kwargs)
        self.benchmark_dir.cleanup()


class DummyContext:
    """Dummy Context class for testing shared state initialization."""

    class params:
        """Dummy param variable."""

        round_timeout_seconds: float = 1.0

    _skill: MagicMock = MagicMock()
    skill_id = "dummy_skill_id"

    @property
    def is_abstract_component(self) -> bool:
        """Mock for is_abstract."""
        return True



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains all the storing operations of the behaviours."""


import json
import os.path
from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import Any, Callable, Dict, Optional, TypeVar, Union, cast

from packages.valory.skills.abstract_round_abci.io_.paths import create_pathdirs


StoredJSONType = Union[dict, list]
NativelySupportedSingleObjectType = StoredJSONType
NativelySupportedMultipleObjectsType = Dict[str, NativelySupportedSingleObjectType]
NativelySupportedObjectType = Union[
    NativelySupportedSingleObjectType, NativelySupportedMultipleObjectsType
]
NativelySupportedStorerType = Callable[[str, NativelySupportedObjectType, Any], None]
CustomObjectType = TypeVar("CustomObjectType")
CustomStorerType = Callable[[str, CustomObjectType, Any], None]
SupportedSingleObjectType = Union[NativelySupportedObjectType, CustomObjectType]
SupportedMultipleObjectsType = Dict[str, SupportedSingleObjectType]
SupportedObjectType = Union[SupportedSingleObjectType, SupportedMultipleObjectsType]
SupportedStorerType = Union[NativelySupportedStorerType, CustomStorerType]
NativelySupportedJSONStorerType = Callable[
    [str, Union[StoredJSONType, Dict[str, StoredJSONType]], Any], None
]


class SupportedFiletype(Enum):
    """Enum for the supported filetypes of the IPFS interacting methods."""

    JSON = auto()


class AbstractStorer(ABC):
    """An abstract `Storer` class."""

    def __init__(self, path: str):
        """Initialize an abstract storer."""
        self._path = path
        # Create the dirs of the path if it does not exist.
        create_pathdirs(path)

    @abstractmethod
    def serialize_object(
        self, filename: str, obj: SupportedSingleObjectType, **kwargs: Any
    ) -> Dict[str, str]:
        """Store a single file."""

    def store(
        self, obj: SupportedObjectType, multiple: bool, **kwargs: Any
    ) -> Dict[str, str]:
        """Serialize one or multiple objects."""
        serialized_files: Dict[str, str] = {}
        if multiple:
            if not isinstance(obj, dict):  # pragma: no cover
                raise ValueError(
                    f"Cannot store multiple files of type {type(obj)}!"
                    f"Should be a dictionary of filenames mapped to their objects."
                )
            for filename, single_obj in obj.items():
                filename = os.path.join(self._path, filename)
                serialized_file = self.serialize_object(filename, single_obj, **kwargs)
                serialized_files.update(**serialized_file)
        else:
            serialized_file = self.serialize_object(self._path, obj, **kwargs)
            serialized_files.update(**serialized_file)
        return serialized_files


class JSONStorer(AbstractStorer):
    """A JSON file storer."""

    def serialize_object(
        self, filename: str, obj: NativelySupportedSingleObjectType, **kwargs: Any
    ) -> Dict[str, str]:
        """
        Serialize an object to JSON.

        :param filename: under which name the provided object should be serialized. Note that it will appear in IPFS with this name.
        :param obj: the object to store.
        :returns: a dict mapping the name to the serialized object.
        """
        if not any(isinstance(obj, type_) for type_ in (dict, list)):
            raise ValueError(  # pragma: no cover
                f"`JSONStorer` cannot be used with a {type(obj)}! Only with a {StoredJSONType}"
            )
        try:
            serialized_object = json.dumps(obj, ensure_ascii=False, indent=4)
            name_to_obj = {filename: serialized_object}
            return name_to_obj
        except (TypeError, OSError) as e:  # pragma: no cover
            raise IOError(str(e)) from e


class Storer(AbstractStorer):
    """Class which serializes objects."""

    def __init__(
        self,
        filetype: Optional[Any],
        custom_storer: Optional[CustomStorerType],
        path: str,
    ):
        """Initialize a `Storer`."""
        super().__init__(path)
        self._filetype = filetype
        self._custom_storer = custom_storer
        self._filetype_to_storer: Dict[Enum, SupportedStorerType] = {
            SupportedFiletype.JSON: cast(
                NativelySupportedJSONStorerType, JSONStorer(path).serialize_object
            ),
        }

    def serialize_object(
        self, filename: str, obj: NativelySupportedObjectType, **kwargs: Any
    ) -> Dict[str, str]:
        """Store a single object."""
        storer = self._get_single_storer_from_filetype()
        return storer(filename, obj, **kwargs)  # type: ignore

    def _get_single_storer_from_filetype(self) -> SupportedStorerType:
        """Get an object storer from a given filetype or keep a custom storer."""
        if self._filetype is not None:
            return self._filetype_to_storer[self._filetype]

        if self._custom_storer is not None:  # pragma: no cover
            return self._custom_storer

        raise ValueError(  # pragma: no cover
            "Please provide either a supported filetype or a custom storing function."
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Analyse dialogue definitions."""


import importlib
import importlib.machinery
import importlib.util
import types
from importlib.machinery import ModuleSpec
from pathlib import Path
from typing import Dict, List, cast

import yaml

from autonomy.analyse.constants import CLASS_NAME, DIALOGUES, DIALOGUES_FILE, MODELS


def load_dialogues_module_from_skill_path(skill_path: Path) -> types.ModuleType:
    """Load `dialogues.py` module for the given skill."""

    loader = importlib.machinery.SourceFileLoader(
        DIALOGUES, str(skill_path / DIALOGUES_FILE)
    )
    spec = cast(ModuleSpec, importlib.util.spec_from_loader(DIALOGUES, loader))
    module = importlib.util.module_from_spec(spec)
    loader.exec_module(module)

    return module


def validate_and_get_dialogues(
    models_configuration: Dict[str, Dict[str, str]]
) -> Dict[str, str]:
    """Returns dialogue names to class name mappings"""
    mapping = {}
    for name, info in models_configuration.items():
        name_ends_with_dialogue = name.endswith(DIALOGUES)

        class_name = info[CLASS_NAME]
        class_ends_with_dialogue = class_name.lower().endswith(DIALOGUES)

        if not class_ends_with_dialogue and not name_ends_with_dialogue:
            continue  # pragma: nocover

        if name_ends_with_dialogue and not class_ends_with_dialogue:
            raise ValueError(
                f"Class name for a dialogue should end with `Dialogues`; dialogue={name}; class={class_name}"
            )

        if class_ends_with_dialogue and not name_ends_with_dialogue:
            raise ValueError(
                f"Dialogue name should end with `dialogues`; dialogue={name}; class={class_name}"
            )

        mapping[name] = class_name

    return mapping


def check_dialogues_in_a_skill_package(config_file: Path, dialogues: List[str]) -> None:
    """Check dialogue definitions."""

    if not (config_file.parent / DIALOGUES_FILE).exists():
        raise FileNotFoundError(f"dialogue file does not exist in {config_file.parent}")

    module = load_dialogues_module_from_skill_path(skill_path=config_file.parent)
    module_attributes = dir(module)

    with open(str(config_file), mode="r", encoding="utf-8") as fp:
        config = yaml.safe_load(fp)
        dialogue_to_class = validate_and_get_dialogues(
            models_configuration=cast(Dict[str, Dict[str, str]], config[MODELS])
        )

        for dialogue_name in dialogues:
            if not dialogue_name.endswith(DIALOGUES):
                dialogue_name = f"{dialogue_name}_{DIALOGUES}"

            if dialogue_name not in dialogue_to_class:
                raise ValueError(
                    f"Common dialogue '{dialogue_name}' is not defined in {config_file}"
                )

            if dialogue_to_class[dialogue_name] not in module_attributes:
                raise ValueError(
                    f"dialogue {dialogue_to_class[dialogue_name]} declared in {config_file} "
                    f"is missing from {config_file.parent / DIALOGUES}.py"
                )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test transaction settlement integration test tool."""

from pathlib import Path
from typing import cast
from unittest import mock

import pytest
from aea.exceptions import AEAActException
from web3.types import Nonce, Wei

from packages.valory.protocols.contract_api import ContractApiMessage
from packages.valory.protocols.ledger_api import LedgerApiMessage
from packages.valory.skills import transaction_settlement_abci
from packages.valory.skills.abstract_round_abci.base import AbciAppDB
from packages.valory.skills.abstract_round_abci.tests.test_tools.base import (
    FSMBehaviourTestToolSetup,
)
from packages.valory.skills.transaction_settlement_abci.behaviours import (
    FinalizeBehaviour,
)
from packages.valory.skills.transaction_settlement_abci.rounds import (
    SynchronizedData as TxSettlementSynchronizedSata,
)
from packages.valory.skills.transaction_settlement_abci.test_tools.integration import (
    _GnosisHelperIntegration,
    _SafeConfiguredHelperIntegration,
    _TxHelperIntegration,
)


DUMMY_TX_HASH = "a" * 234


class Test_SafeConfiguredHelperIntegration(FSMBehaviourTestToolSetup):
    """Test_SafeConfiguredHelperIntegration"""

    test_cls = _SafeConfiguredHelperIntegration

    def test_instantiation(self) -> None:
        """Test instantiation"""

        self.set_path_to_skill()
        self.test_cls.make_ledger_api_connection_callable = (
            lambda *_, **__: mock.MagicMock()
        )
        test_instance = cast(_SafeConfiguredHelperIntegration, self.setup_test_cls())

        assert test_instance.keeper_address in test_instance.safe_owners


class Test_GnosisHelperIntegration(FSMBehaviourTestToolSetup):
    """Test_SafeConfiguredHelperIntegration"""

    test_cls = _GnosisHelperIntegration

    def test_instantiation(self) -> None:
        """Test instantiation"""

        self.set_path_to_skill()
        self.test_cls.make_ledger_api_connection_callable = (
            lambda *_, **__: mock.MagicMock()
        )
        test_instance = cast(_GnosisHelperIntegration, self.setup_test_cls())

        assert test_instance.safe_contract_address
        assert test_instance.gnosis_instance
        assert test_instance.ethereum_api


class Test_TxHelperIntegration(FSMBehaviourTestToolSetup):
    """Test_SafeConfiguredHelperIntegration"""

    test_cls = _TxHelperIntegration

    def instantiate_test(self) -> _TxHelperIntegration:
        """Instantiate the test"""

        path_to_skill = Path(transaction_settlement_abci.__file__).parent
        self.set_path_to_skill(path_to_skill=path_to_skill)
        self.test_cls.make_ledger_api_connection_callable = (
            lambda *_, **__: mock.MagicMock()
        )

        db = AbciAppDB(
            setup_data={"all_participants": [f"agent_{i}" for i in range(4)]}
        )
        self.test_cls.tx_settlement_synchronized_data = TxSettlementSynchronizedSata(db)

        test_instance = cast(_TxHelperIntegration, self.setup_test_cls())
        return test_instance

    def test_sign_tx(self) -> None:
        """Test sign_tx"""

        test_instance = self.instantiate_test()
        test_instance.tx_settlement_synchronized_data.db.update(
            most_voted_tx_hash=DUMMY_TX_HASH
        )

        target = test_instance.gnosis_instance.functions.getOwners
        return_value = test_instance.safe_owners
        with mock.patch.object(
            target, "call", new_callable=lambda: lambda: return_value
        ):
            test_instance.sign_tx()

    def test_sign_tx_failure(self) -> None:
        """Test sign_tx failure"""

        test_instance = self.instantiate_test()
        test_instance.tx_settlement_synchronized_data.db.update(
            most_voted_tx_hash=DUMMY_TX_HASH
        )

        target = test_instance.gnosis_instance.functions.getOwners
        with mock.patch.object(target, "call", new_callable=lambda: lambda: {}):
            with pytest.raises(AssertionError):
                test_instance.sign_tx()

    def test_send_tx(self) -> None:
        """Test send tx"""

        test_instance = self.instantiate_test()

        nonce = Nonce(0)
        gas_price = {"maxPriorityFeePerGas": Wei(0), "maxFeePerGas": Wei(0)}
        behaviour = cast(FinalizeBehaviour, test_instance.behaviour.current_behaviour)
        behaviour.params.mutable_params.gas_price = gas_price
        behaviour.params.mutable_params.nonce = nonce

        contract_api_message = ContractApiMessage(
            performative=ContractApiMessage.Performative.RAW_TRANSACTION,  # type: ignore
            raw_transaction=ContractApiMessage.RawTransaction(
                ledger_id="", body={"nonce": str(nonce), **gas_price}
            ),
        )

        ledger_api_message = LedgerApiMessage(
            performative=LedgerApiMessage.Performative.TRANSACTION_DIGEST,  # type: ignore
            transaction_digest=LedgerApiMessage.TransactionDigest(
                ledger_id="", body=""
            ),
        )

        return_value = contract_api_message, None, ledger_api_message

        with mock.patch.object(
            test_instance,
            "process_n_messages",
            new_callable=lambda: lambda *x, **__: return_value,
        ):
            test_instance.send_tx()

    def test_validate_tx(self) -> None:
        """Test validate_tx"""

        test_instance = self.instantiate_test()
        test_instance.tx_settlement_synchronized_data.db.update(
            tx_hashes_history="a" * 64
        )

        contract_api_message = ContractApiMessage(
            performative=ContractApiMessage.Performative.STATE,  # type: ignore
            state=ContractApiMessage.State(
                ledger_id="",
                body={"verified": True},
            ),
        )
        return_value = None, contract_api_message

        with mock.patch.object(
            test_instance,
            "process_n_messages",
            new_callable=lambda: lambda *x, **__: return_value,
        ):
            test_instance.validate_tx()

    def test_validate_tx_timeout(self) -> None:
        """Test validate_tx timeout"""

        test_instance = self.instantiate_test()
        synchronized_data = test_instance.tx_settlement_synchronized_data
        assert synchronized_data.n_missed_messages == 0
        test_instance.validate_tx(simulate_timeout=True)
        assert synchronized_data.n_missed_messages == 1

    def test_validate_tx_failure(self) -> None:
        """Test validate tx failure"""

        test_instance = self.instantiate_test()

        with pytest.raises(
            AEAActException, match="FSM design error: tx hash should exist"
        ):
            test_instance.validate_tx()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 valory
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test messages module for abci protocol."""

# pylint: disable=too-many-statements,too-many-locals,no-member,too-few-public-methods,redefined-builtin
from typing import List

from aea.test_tools.test_protocol import BaseProtocolMessagesTestCase

from packages.valory.protocols.abci.custom_types import (
    BlockID,
    BlockParams,
    CheckTxType,
    CheckTxTypeEnum,
    ConsensusParams,
    ConsensusVersion,
    Duration,
    Event,
    EventAttribute,
    Events,
    Evidence,
    EvidenceParams,
    EvidenceType,
    Evidences,
    Header,
    LastCommitInfo,
    PartSetHeader,
    ProofOp,
    ProofOps,
    PublicKey,
    Result,
    ResultType,
    SnapShots,
    Snapshot,
    Timestamp,
    Validator,
    ValidatorParams,
    ValidatorUpdate,
    ValidatorUpdates,
    VersionParams,
    VoteInfo,
)
from packages.valory.protocols.abci.message import AbciMessage


event = Event("type", [EventAttribute(b"key", b"value", True)])
events = Events([event, event])

header = Header(
    ConsensusVersion(0, 0),
    "chain_id",
    0,
    Timestamp(0, 0),
    BlockID(b"hash", PartSetHeader(0, b"hash")),
    b"last_commit_hash",
    b"data_hash",
    b"validators_hash",
    b"next_validators_hash",
    b"consensus_hash",
    b"app_hash",
    b"last_results_hash",
    b"evidence_hash",
    b"proposer_address",
)

validator = Validator(b"address", 0)
last_commit_info = LastCommitInfo(
    0,
    [
        VoteInfo(validator, True),
        VoteInfo(validator, False),
    ],
)

consensus_params = ConsensusParams(
    BlockParams(0, 0),
    EvidenceParams(0, Duration(0, 0), 0),
    ValidatorParams(["pub_key"]),
    VersionParams(0),
)

evidences = Evidences(
    [
        Evidence(EvidenceType.UNKNOWN, validator, 0, Timestamp(0, 0), 0),
        Evidence(EvidenceType.DUPLICATE_VOTE, validator, 0, Timestamp(0, 0), 0),
    ]
)


validators = ValidatorUpdates(
    [
        ValidatorUpdate(
            PublicKey(b"pub_key_bytes", PublicKey.PublicKeyType.ed25519), 1
        ),
        ValidatorUpdate(
            PublicKey(b"pub_key_bytes", PublicKey.PublicKeyType.secp256k1), 2
        ),
    ]
)
snapshot = Snapshot(0, 0, 0, b"hash", b"metadata")
snapshots = SnapShots([snapshot, snapshot])
proof_ops = (
    ProofOps(
        [
            ProofOp("type", b"key", b"data"),
            ProofOp("type", b"key", b"data"),
        ]
    ),
)


class TestMessageAbci(BaseProtocolMessagesTestCase):
    """Test for the 'abci' protocol message."""

    MESSAGE_CLASS = AbciMessage

    def build_messages(self) -> List[AbciMessage]:  # type: ignore[override]
        """Build the messages to be used for testing."""
        return [
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_ECHO,
                message="some str",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_FLUSH,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_INFO,
                version="some str",
                block_version=12,
                p2p_version=12,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_SET_OPTION,
                option_key="some str",
                option_value="some str",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_INIT_CHAIN,
                time=Timestamp(seconds=1, nanos=1),
                chain_id="some str",
                consensus_params=consensus_params,
                validators=validators,
                app_state_bytes=b"some_bytes",
                initial_height=12,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_QUERY,
                query_data=b"some_bytes",
                path="some str",
                height=12,
                prove=True,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_BEGIN_BLOCK,
                hash=b"some_bytes",
                header=header,
                last_commit_info=last_commit_info,
                byzantine_validators=evidences,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_CHECK_TX,
                tx=b"some_bytes",
                type=CheckTxType(CheckTxTypeEnum.NEW),
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_DELIVER_TX,
                tx=b"some_bytes",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_END_BLOCK,
                height=12,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_COMMIT,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_LIST_SNAPSHOTS,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_OFFER_SNAPSHOT,
                snapshot=Snapshot(0, 0, 0, b"hash", b"metadata"),
                app_hash=b"some_bytes",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_LOAD_SNAPSHOT_CHUNK,
                height=12,
                format=12,
                chunk_index=12,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_APPLY_SNAPSHOT_CHUNK,
                index=12,
                chunk=b"some_bytes",
                chunk_sender="some str",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_EXCEPTION,
                error="some str",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_ECHO,
                message="some str",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_FLUSH,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_INFO,
                info_data="some str",
                version="some str",
                app_version=12,
                last_block_height=12,
                last_block_app_hash=b"some_bytes",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_SET_OPTION,
                code=12,
                log="some str",
                info="some str",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_INIT_CHAIN,
                consensus_params=consensus_params,
                validators=validators,
                app_hash=b"some_bytes",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_QUERY,
                code=12,
                log="some str",
                info="some str",
                index=12,
                key=b"some_bytes",
                value=b"some_bytes",
                proof_ops=ProofOps(
                    [
                        ProofOp("type", b"key", b"data"),
                        ProofOp("type", b"key", b"data"),
                    ]
                ),
                height=12,
                codespace="some str",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_BEGIN_BLOCK,
                events=events,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_CHECK_TX,
                code=12,
                data=b"some_bytes",
                log="some str",
                info="some str",
                gas_wanted=12,
                gas_used=12,
                events=events,
                codespace="some str",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_DELIVER_TX,
                code=12,
                data=b"some_bytes",
                log="some str",
                info="some str",
                gas_wanted=12,
                gas_used=12,
                events=events,
                codespace="some str",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_END_BLOCK,
                validator_updates=validators,
                consensus_param_updates=consensus_params,
                events=events,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_COMMIT,
                data=b"some_bytes",
                retain_height=12,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_LIST_SNAPSHOTS,
                snapshots=snapshots,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_OFFER_SNAPSHOT,
                result=Result(ResultType.UNKNOWN),
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_LOAD_SNAPSHOT_CHUNK,
                chunk=b"some_bytes",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_APPLY_SNAPSHOT_CHUNK,
                result=Result(ResultType.UNKNOWN),
                refetch_chunks=(12,),
                reject_senders=("some str",),
            ),
            AbciMessage(
                performative=AbciMessage.Performative.DUMMY,
                dummy_consensus_params=consensus_params,
            ),
        ]

    def build_inconsistent(self) -> List[AbciMessage]:  # type: ignore[override]
        """Build inconsistent messages to be used for testing."""
        return [
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_ECHO,
                # skip content: message
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_INFO,
                # skip content: version
                block_version=12,
                p2p_version=12,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_SET_OPTION,
                # skip content: option_key
                option_value="some str",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_INIT_CHAIN,
                # skip content: time,app_state_bytes
                chain_id="some str",
                consensus_params=consensus_params,
                validators=validators,
                initial_height=12,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_QUERY,
                # skip content: query_data
                path="some str",
                height=12,
                prove=True,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_BEGIN_BLOCK,
                # skip content: hash
                header=header,
                last_commit_info=last_commit_info,
                byzantine_validators=evidences,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_CHECK_TX,
                # skip content: tx
                type=CheckTxType(CheckTxTypeEnum.NEW),
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_DELIVER_TX,
                # skip content: tx
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_END_BLOCK,
                # skip content: height
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_OFFER_SNAPSHOT,
                # skip content: snapshot
                app_hash=b"some_bytes",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_LOAD_SNAPSHOT_CHUNK,
                # skip content: height
                format=12,
                chunk_index=12,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.REQUEST_APPLY_SNAPSHOT_CHUNK,
                # skip content: index
                chunk=b"some_bytes",
                chunk_sender="some str",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_EXCEPTION,
                # skip content: error
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_ECHO,
                # skip content: message
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_INFO,
                # skip content: info_data
                version="some str",
                app_version=12,
                last_block_height=12,
                last_block_app_hash=b"some_bytes",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_SET_OPTION,
                # skip content: code
                log="some str",
                info="some str",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_INIT_CHAIN,
                # skip content: consensus_params, validators
                app_hash=b"some_bytes",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_QUERY,
                # skip content: code
                log="some str",
                info="some str",
                index=12,
                key=b"some_bytes",
                value=b"some_bytes",
                proof_ops=proof_ops,
                height=12,
                codespace="some str",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_BEGIN_BLOCK,
                # skip content: events
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_CHECK_TX,
                # skip content: code
                data=b"some_bytes",
                log="some str",
                info="some str",
                gas_wanted=12,
                gas_used=12,
                events=events,
                codespace="some str",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_DELIVER_TX,
                # skip content: code
                data=b"some_bytes",
                log="some str",
                info="some str",
                gas_wanted=12,
                gas_used=12,
                events=events,
                codespace="some str",
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_END_BLOCK,
                # skip content: validator_updates
                consensus_param_updates=consensus_params,
                events=events,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_COMMIT,
                # skip content: data
                retain_height=12,
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_LIST_SNAPSHOTS,
                # skip content: snapshots
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_OFFER_SNAPSHOT,
                # skip content: result
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_LOAD_SNAPSHOT_CHUNK,
                # skip content: chunk
            ),
            AbciMessage(
                performative=AbciMessage.Performative.RESPONSE_APPLY_SNAPSHOT_CHUNK,
                # skip content: result
                refetch_chunks=(12,),
                reject_senders=("some str",),
            ),
            AbciMessage(
                performative=AbciMessage.Performative.DUMMY,
                # skip content: dummy_consensus_params
            ),
        ]
